#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List

import numpy as np

from ase import Atoms
from ase import data, units
from ase.neighborlist import NeighborList, natural_cutoffs

from .move import MoveOperator


class SwapOperator(MoveOperator):

    name: str = "swap"

    def __init__(
        self,
        particles: List[str],
        region: dict = {},
        temperature: float = 300,
        pressure: float = 1,
        covalent_ratio=[0.8, 2.0],
        use_rotation: bool = True,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(
            particles=particles,
            region=region,
            temperature=temperature,
            pressure=pressure,
            covalent_ratio=covalent_ratio,
            use_rotation=use_rotation,
            *args,
            **kwargs,
        )

        # NOTE: Prohibit swapping the same type of particles.
        assert (
            len(set(self.particles)) == 2
        ), f"f{self.__class__.__name__} needs two types of particles."

        return

    def run(self, atoms: Atoms, rng=np.random) -> Atoms:
        """"""
        super().run(atoms)

        # - basic
        curr_atoms = atoms
        cell = curr_atoms.get_cell(complete=True)

        # -- neighbour list
        nl = NeighborList(
            self.covalent_max * np.array(natural_cutoffs(curr_atoms)),
            skin=0.0,
            self_interaction=False,
            bothways=True,
        )

        # - swap the species
        for i in range(self.MAX_RANDOM_ATTEMPTS):
            # -- swap
            curr_atoms = atoms.copy()

            # -- pick an atom
            #   either index of an atom or tag of an moiety
            first_pick = self._select_species(curr_atoms, [self.particles[0]], rng=rng)
            second_pick = self._select_species(curr_atoms, [self.particles[1]], rng=rng)
            self._print(f"first: {first_pick} second: {second_pick}")

            # -- find tag atoms
            first_species = curr_atoms[first_pick]  # default copy
            second_species = curr_atoms[second_pick]
            # TODO: deal with pbc
            first_cop = np.average(copy.deepcopy(first_species.get_positions()), axis=0)
            second_cop = np.average(
                copy.deepcopy(second_species.get_positions()), axis=0
            )

            self._print(f"origin: {first_species.symbols} {first_cop}")
            self._print(f"origin: {second_species.symbols} {second_cop}")

            # -- rotate and swap
            first_species = self._rotate_species(first_species, rng=rng)
            second_species = self._rotate_species(second_species, rng=rng)

            curr_atoms.positions[first_pick] += second_cop - first_cop
            curr_atoms.positions[second_pick] += first_cop - second_cop

            first_species = curr_atoms[first_pick]
            second_species = curr_atoms[second_pick]
            # TODO: deal with pbc
            first_cop = np.average(copy.deepcopy(first_species.get_positions()), axis=0)
            second_cop = np.average(
                copy.deepcopy(second_species.get_positions()), axis=0
            )

            self._print(f"swapped: {first_species.symbols} {first_cop}")
            self._print(f"swapped: {second_species.symbols} {second_cop}")

            # -- use neighbour list
            idx_pick = []
            idx_pick.extend(first_pick)
            idx_pick.extend(second_pick)
            if not self.check_overlap_neighbour(nl, curr_atoms, cell, idx_pick):
                self._print(f"succeed to random after {i+1} attempts...")
                break
        else:
            curr_atoms = None

        return curr_atoms

    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()
        params["particles"] = self.particles

        return params

    def __repr__(self) -> str:
        """"""
        content = f"@Modifier {self.__class__.__name__}\n"
        content += (
            f"temperature {self.temperature} [K] pressure {self.pressure} [bar]\n"
        )
        content += "covalent ratio: \n"
        content += f"  min: {self.covalent_min} max: {self.covalent_max}\n"
        content += f"swapped groups: \n"
        content += f"  {self.particles[0]} <-> {self.particles[1]}\n"

        return content


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import Callable, List, Tuple

import networkx as nx

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

from .. import CustomTimer
from .. import StruGraphCreator, extract_chem_envs
from .. import unpack_node_name
from .. import get_unique_environments_based_on_bonds

from .modifier import GraphModifier, DEFAULT_GRAPH_PARAMS
from ..group import create_a_group


def single_remove_adsorbate(
    species: str,
    graph_params: dict,
    target_group: List[dict],
    atoms: Atoms,
    print_func: Callable = print,
    debug_func: Callable = print,
) -> Tuple[List[Atoms], List[nx.Graph]]:
    """Remove selected particles from the structure.

    Currently, only single atom can be removed. TODO: molecule.

    Args:
        graph_params: Parameters for creating graphs.
        spec_params: Parameters for finding species to remove.

    """
    # - create graph from structure
    stru_creator = StruGraphCreator(**graph_params)

    # - check if spec_indices are all species
    natoms = len(atoms)
    group_indices = list(range(natoms))
    for command in target_group:
        curr_indices = create_a_group(atoms, command)
        group_indices = [i for i in group_indices if i in curr_indices]
    debug_func(f"group_indices to remove {group_indices}")

    # TODO: tags for molecule?
    chemical_symbols = atoms.get_chemical_symbols()
    for i in group_indices:
        if chemical_symbols[i] != species:
            raise RuntimeError(
                "Species to remove is inconsistent for those by indices."
            )

    # - get chem envs
    graph = stru_creator.generate_graph(atoms, ads_indices=group_indices)
    chem_envs = extract_chem_envs(
        graph, atoms, group_indices, stru_creator.graph_radius
    )

    # NOTE: for single atom adsorption,
    assert len(chem_envs) == len(
        group_indices
    ), "Single atoms group into one adsorbate. Try reducing the covalent radii."
    # TODO: for molecule adsorption

    # - find unique sites to remove for this structure
    unique_indices = get_unique_environments_based_on_bonds(chem_envs)
    unique_envs = [chem_envs[i] for i in unique_indices]

    # - create sctructures
    unique_frames = []
    for g in unique_envs:
        for u, d in g.nodes.data():
            if d["central_ads"]:
                chem_sym, idx, offset = unpack_node_name(u)
                if chem_sym == species:
                    new_atoms = atoms.copy()
                    del new_atoms[idx]
                    unique_frames.append(new_atoms)
                    break
        else:
            # no valid adsorbate for this structure
            ...

    return unique_frames, unique_envs


class GraphRemoveModifier(GraphModifier):

    name: str = "graph_remove"

    def __init__(
        self,
        species,
        spectators: List[str],
        target_group: List[dict],
        substrates=None,
        graph: dict = DEFAULT_GRAPH_PARAMS,
        *args,
        **kwargs,
    ):
        """Insert an adsorbate on sites according to graph representation."""
        super().__init__(substrates=substrates, *args, **kwargs)

        self.species = species  # make this a node

        self.target_group = target_group

        self.spectators = spectators
        self.graph_params = graph

        # self.check_site_unique = True
        # adsorbate_indices
        # site_radius
        # region

        return

    def _irun(self, substrates: List[Atoms]) -> List[Atoms]:
        """Remove atoms/molecules/adsorbates."""
        self._print("---run remove---")
        graph_params = copy.deepcopy(self.graph_params)
        adsorbate_elements = copy.deepcopy(self.spectators)
        graph_params.update(adsorbate_elements=adsorbate_elements)

        # - get chem envs of selected species that may be removed
        with CustomTimer(name="remove-adsorbate", func=self._print):
            ret = Parallel(n_jobs=self.njobs)(
                delayed(single_remove_adsorbate)(
                    self.species, graph_params, self.target_group, a,
                    print_func=self._print, debug_func=self._debug
                )
                for idx, a in enumerate(substrates)
            )

            ret_frames, ret_envs = [], []
            for i, (frames, envs) in enumerate(ret):
                nenvs = len(envs)
                # TODO: add info since it may be lost in atoms.copy() function
                # for a in frames:
                #    a.info["subid"] = subid
                # -- add data
                ret_envs.extend(envs)
                ret_frames.extend(frames)
                self._print(f"number of sites {nenvs} to remove for substrate {i}.")
        # nsites = len(ret_frames)
        # self._print(f"Total number of chemical environments: {nsites}")

        # - further unique envs among different substrates
        #   only compare chemical environments
        # unique_indices = get_unique_environments_based_on_bonds(ret_envs)
        # created_frames = [ret_frames[i] for i in unique_indices]

        # - compare the graph of chemical environments in the structure
        #   NOTE: if O atoms were to remove, the chem envs of the rest O atoms
        #         are used to compare the structure difference.
        write(self.directory / f"possible_frames.xyz", ret_frames)

        # - get unique structures among substrates
        created_frames = self._compare_structures(
            ret_frames, graph_params, self.target_group
        )

        return created_frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List
from pathlib import Path

import numpy as np

from ase import Atoms

from .selector import AbstractSelector


class InvariantSelector(AbstractSelector):
    """Perform an invariant selection."""

    name = "invariant"

    default_parameters = dict()

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        return

    def _mark_structures(self, frames: List[Atoms], *args, **kwargs) -> None:
        """Return selected indices."""

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import itertools
import multiprocessing

from typing import Optional, List, Tuple

import numpy as np

import joblib

from ase import Atoms
from ase.data import atomic_numbers
from ase.geometry import find_mic

from . import registers
from .builder import StructureModifier
from .utils import (
    check_overlap_neighbour,
    convert_composition_to_list,
    rotate_a_molecule,
)


RANDOM_INTEGER_HIGH: int = 1e16


def insert_species(
    substrate: Atoms,
    composition_list,
    region,
    intermol_dismin,
    covalent_ratio,
    custom_dmin_dict,
    random_state,
) -> Optional[Atoms]:
    """"""
    rng = np.random.Generator(np.random.PCG64(random_state))
    # rng = random_state

    atoms = substrate
    atoms.set_tags(0)

    excluded_pairs = list(itertools.permutations(range(len(atoms)), 2))

    species_list = itertools.chain(
        *[[k for i in range(v)] for k, v in composition_list]
    )
    species_list = sorted(species_list, key=lambda a: a.get_chemical_formula())
    num_species = len(species_list)

    random_positions = region.get_random_positions(size=num_species, rng=rng)
    if num_species > 1:
        pair_positions = np.array(list(itertools.combinations(random_positions, 2)))
        raw_vectors = pair_positions[:, 0, :] - pair_positions[:, 1, :]
        mic_vecs, mic_dis = find_mic(v=raw_vectors, cell=atoms.cell)
        if np.min(mic_dis) >= intermol_dismin:
            is_molecule_valid = True
        else:
            is_molecule_valid = False
    else:
        is_molecule_valid = True

    if is_molecule_valid:
        intra_bonds = []
        for a, p in zip(species_list, random_positions):
            # count number of atoms
            prev_num_atoms = len(atoms)
            curr_num_atoms = prev_num_atoms + len(a)
            intra_bonds.extend(
                list(itertools.permutations(range(prev_num_atoms, curr_num_atoms), 2))
            )
            # rotate and translate
            a = rotate_a_molecule(a, use_com=True, rng=rng)
            # a.translate(p - np.mean(a.positions, axis=0))
            a.translate(p - a.get_center_of_mass())
            a.set_tags(int(np.max(atoms.get_tags()) + 1))
            atoms += a

        excluded_pairs.extend(intra_bonds)
        if not check_overlap_neighbour(
            atoms,
            covalent_ratio=covalent_ratio,
            custom_dmin_dict=custom_dmin_dict,
            excluded_pairs=excluded_pairs,
        ):
            atoms = None
        else:
            ...
    else:
        atoms = None

    return atoms


class InsertModifier(StructureModifier):

    name = "insert"

    def __init__(
        self,
        region,
        composition: dict,
        custom_dmin: list = [],
        covalent_ratio=[0.8, 2.0],
        molecular_distol=[-np.inf, np.inf],
        max_times_size: int = 5,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(*args, **kwargs)

        # - system definition
        self.region = copy.deepcopy(region)  # TODO: str, dict, or a Region obeject?
        shape = region.pop("method", None)
        self._region = registers.create("region", shape, convert_name=True, **region)

        self.composition = composition
        self._composition_list = convert_composition_to_list(composition, self._region)

        for k, v in self._composition_list:
            if len(k) > 1:
                self._insert_molecule = True
                break
        else:
            self._insert_molecule = False

        self.intermol_dismin = molecular_distol[0]
        self.intermol_dismax = molecular_distol[1]

        custom_dmin_dict = {}
        for i, j, d in custom_dmin:
            s_i, s_j = atomic_numbers[i], atomic_numbers[j]
            custom_dmin_dict[(s_i, s_j)] = d
            custom_dmin_dict[(s_j, s_i)] = d
        self.custom_dmin_dict = custom_dmin_dict

        # - bond distance check
        self.covalent_ratio = covalent_ratio
        self.MAX_TIMES_SIZE = max_times_size

        return

    def run(
        self, substrates: List[Atoms] = None, size: int = 1, *args, **kwargs
    ) -> List[Atoms]:
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        # If no substrates, create a empty box with the region as its cell TODO:
        _substrates = self.substrates
        if _substrates is None:
            _substrates = [Atoms()]

        num_substrates = len(_substrates)

        self._print(f"{self.njobs =}")

        # PERF: For easy random tasks, use stratified parallel run.
        #       try small_times_size first and increase it if not
        #       enough structures are generated.
        num_attempts = 0
        combined_frames = [[] for _ in range(num_substrates)]
        num_frames = 0
        for i in range(self.MAX_TIMES_SIZE):
            curr_max_attempts = self.njobs * 2 ** int(
                np.log(num_substrates * size - num_frames)
            )
            num_attempts += curr_max_attempts*num_substrates
            ret = self._irun(_substrates, max_attempts=curr_max_attempts)

            for batch_frames in ret:
                for i, atoms in enumerate(batch_frames):
                    if len(combined_frames[i]) < size:
                        if atoms is not None:
                            combined_frames[i].append(atoms)
                    else:
                        ...
                combined_num_frames = [len(cf) for cf in combined_frames]
                if np.all([cnf == size for cnf in combined_num_frames]):
                    break

            combined_num_frames = [len(cf) for cf in combined_frames]
            num_frames = np.sum(combined_num_frames)
            self._print(
                f"Need {size}*{num_substrates} structures and {num_frames} is created in {curr_max_attempts} attempts."
            )
            if np.all([cnf == size for cnf in combined_num_frames]):
                break

        if num_attempts >= RANDOM_INTEGER_HIGH:
            self._print("The random structures may have duplicates.")

        frames = list(itertools.chain(*combined_frames))
        num_frames = len(frames)
        self._print(f"{num_frames =}")

        if num_frames != size * num_substrates:
            raise RuntimeError(
                f"Need {size}*{num_substrates} structures but only {num_frames} is created."
            )

        return frames

    def _irun(self, substrates, max_attempts: int) -> List[List[Atoms]]:
        """"""
        # prepare inputs for parallel
        batches = []
        for _ in range(max_attempts):
            prepared_substrates = [copy.deepcopy(a) for a in substrates]
            # print(f"{prepared_substrates =}")

            num_prepared_substrates = len(prepared_substrates)
            prepared_random_states = self.rng.integers(
                low=0, high=RANDOM_INTEGER_HIGH, size=num_prepared_substrates
            )
            batches.append([prepared_substrates, prepared_random_states])

        # if multiprocessing.get_start_method() != "spawn":
        if True:
            backend = "loky"
            ret = joblib.Parallel(n_jobs=self.njobs, backend=backend)(
                joblib.delayed(insert_species_batch)(
                    substrates=curr_substrates,
                    composition_list=self._composition_list,
                    region=self._region,
                    intermol_dismin=self.intermol_dismin,
                    covalent_ratio=self.covalent_ratio,
                    custom_dmin_dict=self.custom_dmin_dict,
                    random_states=curr_random_states,
                )
                for curr_substrates, curr_random_states in batches
            )
        else:
            raise RuntimeError("multiprocessing.")

        return ret


def insert_species_batch(
    substrates: Atoms,
    composition_list,
    region,
    intermol_dismin,
    covalent_ratio,
    custom_dmin_dict,
    random_states,
):
    """"""
    frames = []
    for atoms, random_state in zip(substrates, random_states):
        # print(atoms)
        new_atoms = insert_species(
            substrate=atoms,
            composition_list=composition_list,
            region=region,
            intermol_dismin=intermol_dismin,
            covalent_ratio=covalent_ratio,
            custom_dmin_dict=custom_dmin_dict,
            random_state=random_state,
        )
        frames.append(new_atoms)

    return frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
from typing import NoReturn, Union, List, Mapping

import numpy as np

from ase import Atoms
from ase.io import read, write

from dscribe.descriptors import SOAP

from .selector import AbstractSelector
from .cur import cur_selection, fps_selection


"""Selector using descriptors.
"""


class DescriptorSelector(AbstractSelector):
    """Selector using descriptors."""

    name = "dscribe"

    default_parameters = dict(
        mode="stru",
        descriptor=None,
        sparsify=dict(
            # -- cur
            method="cur",
            zeta="-1",
            strategy="descent",
            # -- fps
            # method = "fps",
            # min_distance = 0.1,
            # metric = "euclidean",
            # metric_params = {}
        ),
        number=[4, 0.2],
        verbose=False,
    )

    def __init__(self, directory="./", *args, **kwargs):
        """"""
        super().__init__(directory=directory, *args, **kwargs)

        # - check params
        criteria_method = self.sparsify["method"]
        assert criteria_method in [
            "cur",
            "fps",
        ], f"Unknown selection method {criteria_method}."

        assert self.mode in ["stru", "traj"], f"Unknown selection mode {self.mode}."

        return

    def _compute_descripter(self, frames: List[Atoms]) -> np.array:
        """Calculate vector-based descriptors.

        Each structure is represented by a vector.

        """
        features_path = self.directory / "features.npy"

        self._print("start calculating features...")
        desc_params = copy.deepcopy(self.descriptor)
        desc_name = desc_params.pop("name", None)

        features = None
        if desc_name == "soap":
            soap = SOAP(**desc_params)
            ndim = soap.get_number_of_features()
            self._print(f"soap descriptor dimension: {ndim}")
            features = soap.create(frames, n_jobs=self.njobs)
        else:
            raise RuntimeError(f"Unknown descriptor {desc_name}.")
        self._print("finished calculating features...")

        # - save calculated features
        features = features.reshape(-1, ndim)
        if self.verbose:
            np.save(features_path, features)
            self._print(f"number of soap instances {len(features)}")

        return features

    def _mark_structures(self, data, *args, **kwargs) -> None:
        """Mark structures.

        The selected_indices is the local indices for input markers.

        """
        # - group markers
        if self.axis is None:
            marker_groups = dict(all=data.markers)
        else:
            marker_groups = {}
            for k, v in itertools.groupby(data.markers, key=lambda x: x[self.axis]):
                if k in marker_groups:
                    marker_groups[k].extend(list(v))
                else:
                    marker_groups[k] = list(v)
        self._debug(f"marker_groups: {marker_groups}")

        selected_markers = []
        features, sind_grps, oind_grps = None, {}, {}
        for grp_name, markers in marker_groups.items():
            frames = data.get_marked_structures(markers)  # reference of atoms

            # -
            curr_features, curr_selected_indices = self._select_structures(frames)
            # - update markers
            curr_selected_markers = [markers[i] for i in curr_selected_indices]
            selected_markers.extend(curr_selected_markers)

            # - prepare for plot
            if curr_selected_indices:
                if features is None:
                    curr_nframes = 0
                    features = curr_features
                else:
                    curr_nframes = features.shape[0]
                    features = np.vstack((features, curr_features))
                # - selected ones
                sind_grps[grp_name] = [x + curr_nframes for x in curr_selected_indices]
                # - other ones
                oind_grps[grp_name] = [x + curr_nframes for x in range(len(markers))]

        if any([len(v) for k, v in sind_grps.items()]):
            self._plot_results(features, sind_grps, oind_grps)

        data.markers = np.array(selected_markers)

        return

    def _select_structures(self, frames: List[Atoms]):
        """"""
        nframes = len(frames)
        num_fixed = self._parse_selection_number(nframes)

        # NOTE: currently, only cur and fps are supported
        # TODO: clustering ...
        if num_fixed > 0:
            features = self._compute_descripter(frames)
            if nframes == 1:
                scores, selected_indices = [np.NaN], [0]
            else:
                scores, selected_indices = self._sparsify(features, num_fixed)
        else:
            features, scores, selected_indices = None, [], []

        # - add score to atoms
        #   only save scores from last property
        for score, i in zip(scores, selected_indices):
            frames[i].info["score"] = score

        return features, selected_indices

    def _sparsify(self, features, num_fixed: int):
        """"""
        # TODO: sparsify each traj separately?
        criteria_params = copy.deepcopy(self.sparsify)
        method = criteria_params.pop("method", "cur")
        if method == "cur":
            # -- cur decomposition
            scores, selected_indices = cur_selection(
                features, num_fixed, **criteria_params, rng=self.rng
            )
        elif method == "fps":
            scores, selected_indices = fps_selection(
                features, num_fixed, **criteria_params, rng=self.rng
            )
        else:
            ...

        return scores, selected_indices

    def _plot_results(self, features, groups: dict, others: dict, *args, **kwargs):
        """"""
        # - plot selection
        from sklearn.decomposition import PCA
        import matplotlib.pyplot as plt

        try:
            plt.style.use("presentation")
        except Exception as e:
            ...

        if features.shape[0] > 1:
            reducer = PCA(n_components=2)
            reducer.fit(features)
            proj = reducer.transform(features)

            fig, ax = plt.subplots(1, 1, figsize=(12, 8))

            for grp_name, inds in groups.items():
                sc = ax.scatter(
                    proj[others[grp_name], 0],
                    proj[others[grp_name], 1],
                    marker="o",
                    alpha=0.25,
                    label=f"grp-{grp_name} {len(others[grp_name])} -> {len(inds)}",
                )
                # --
                selected_proj = reducer.transform(np.array([features[i] for i in inds]))
                ax.scatter(
                    selected_proj[:, 0],
                    selected_proj[:, 1],
                    marker="*",
                    alpha=0.5,
                    color="r",
                )
            ax.legend(fontsize=12)
            ax.axis("off")
            fig.savefig(self.info_fpath.parent / (self.info_fpath.stem + ".png"))
            plt.close()
        else:
            ...  # Cannot plot PCA with only one structure...

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

from ase.calculators.calculator import Calculator


class TimeIOCalculator(Calculator):

    def __init__(self, pace: int=1, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        self.pace = pace

        self._num_steps = 0

        return

    @property
    def num_steps(self) -> int:
        """Finished steps that match the host driver e.g. MD."""

        return self._num_steps

    @property
    def log_fpath(self):
        """"""

        return pathlib.Path(self.directory) / "calc.log"

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=["positions", "numbers", "cell"],
    ):
        super().calculate(atoms, properties, system_changes)

        if self.num_steps == 0:
            self._write_first_step()

        self.results, self.step_info = self._icalculate(atoms, properties, system_changes)

        if self.num_steps % self.pace == 0:
            self._write_step()

        self._num_steps += 1

        return

    def _icalculate(self, atoms, properties, system_changes):
        """"""

        raise NotImplementedError()

    def _write_first_step(self):
        """"""

        raise NotImplementedError()

    def _write_step(self):
        """"""

        raise NotImplementedError()

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import collections
import itertools
from typing import List

import numpy as np
import scipy as sp 
from scipy import optimize

import matplotlib.pyplot as plt

from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs

from itertools import permutations, product

from gdpx.potential.trainer import AbstractTrainer
from .gptools import compute_body3_descriptor

def switch_function_value_and_grad(r, r_cut):
    """Switch function that smoothes distance to cutoff.
    """

    return (r-r_cut)**2, 2*(r-r_cut)

#def switch_function_value_and_grad(r, r_cut):
#    """Switch function that smoothes distance to cutoff.
#    """
#
#    return np.array([[1.]]), np.array([[0.]])

def switch_function_value_and_grad_b3(r, r_cut):
    """Switch function that smoothes distance to cutoff.
    """
    r_diff = (r-r_cut)
    v = np.prod(r_diff**2, axis=1)[:, np.newaxis]

    return v, 2./r_diff*v

#def switch_function_value_and_grad_b3(r, r_cut):
#    """Switch function that smoothes distance to cutoff.
#    """
#
#    return np.array([[1.]]), np.array([[0.]])

def gaussian_kernel(x1, x2, delta=0.2, theta=0.5):
    """
        x1: shape (num_points, num_feature_dim)
        x2: shape (num_sparse_points, num_feature_dim)

        -> shape (num_points, num_sparse_points)

    """
    # shape (num_points, num_sparse_points, num_feature_dim)
    x_diff = x1[:, np.newaxis, :].repeat(x2.shape[0], axis=1) - x2[np.newaxis, :, :]

    return delta**2*np.exp(-np.sum(x_diff**2, axis=-1)/2./theta**2)

def gaussian_kernel_value_and_grad(x1, x2, delta=0.2, theta=0.5):
    """
    """
    # shape (num_points, num_sparse_points, num_feature_dim)
    x_diff = x1 - x2.T
    #x_diff = np.linalg.norm((x1 - x2.T)[:, :, np.newaxis], axis=-1)
    v = delta**2*np.exp(-x_diff**2/2./theta**2)

    # gradient wrt x1
    g0 = v*(-x_diff/theta**2) # <---

    # grad wrt delta
    g2 = 2*v/delta

    # grad wrt theta
    g3 = v*(x_diff**2/theta**3)

    # gradient wrt x1 wrt theta
    #g0theta = ((2*x_diff/theta**3) + (-x_diff**3/theta**5))/(-x_diff/theta**2)
    #g0theta = ((-2/theta) + (x_diff**2/theta**3))

    return (v, g0, g2, g3)

def xxx_gaussian_kernel_value_and_grad(x1, x2, delta=0.2, theta=0.5):
    """
        x1: shape (num_points, num_feature_dim)
        x2: shape (num_sparse_points, num_feature_dim)

        v: shape (num_points, num_sparse_points)
        v: shape (num_points, num_sparse_points)
        ### g: shape (num_points, num_sparse_points, num_feature_dim)

    """
    #x_diff = x1 - x2
    #v = delta**2*np.exp(-x_diff**2/2./theta**2)

    # shape (num_points, num_sparse_points, num_feature_dim)
    x_diff = x1[:, np.newaxis, :].repeat(x2.shape[0], axis=1) - x2

    v = delta**2*np.exp(-np.sum(x_diff**2, axis=-1)/2./theta**2)

    # gradient with respect to x_diff
    #g = v*(-np.linalg.norm(x_diff, axis=-1)/theta**2) 

    num_feature_dim = x_diff.shape[-1]
    g = (
        np.repeat(v[:, :, np.newaxis], num_feature_dim, axis=-1) * 
        (-x_diff/theta**2)
    ).squeeze(axis=-1)

    return (v, g)
    
def compute_body2_descriptor(frames, r_cut: float):
    """"""
    # -- get neighbours
    distance_mapping_list = []
    distances = []
    dis_derivs = []

    for i_frame, atoms in enumerate(frames):
        natoms = len(atoms)
        nl = NeighborList(
            cutoffs=[r_cut/2.]*natoms, skin=0.0, sorted=False,
            self_interaction=False, bothways=True
        )
        nl.update(atoms)
        for i in range(natoms):
            nei_indices, nei_offsets = nl.get_neighbors(i)
            #print(i)
            #print(nei_indices, nei_offsets)
            for j, offset in zip(nei_indices, nei_offsets):
                pos_i, pos_j = atoms.positions[i, :], atoms.positions[j, :]
                pos_vec = pos_i - pos_j
                distance = np.linalg.norm(
                    pos_vec + np.dot(offset, atoms.get_cell())
                )
                distances.append(distance)
                dis_deriv = np.hstack([pos_vec/distance, -pos_vec/distance])
                dis_derivs.append(dis_deriv)
                distance_mapping_list.append((i_frame, i, j))

    # distance descriptor, shape (num_distances, 1)
    distances = np.array(distances)[:, np.newaxis] 
    num_distances = distances.shape[0]
    assert len(distance_mapping_list) == num_distances

    dis_derivs = np.array(dis_derivs)

    return distances, distance_mapping_list, dis_derivs
    
def compute_distance_kernal_matrices(
    delta, sigma2, nframes, natoms_list, y_data, 
    distances, dis_derivs, distance_mapping_list, sparse_points
):
    """"""
    # --
    natoms_tot = np.sum(natoms_list)
    num_sparse = sparse_points.shape[0]
    num_points = y_data.shape[0]

    # -- compute Kmm
    Kmm, _, Kmm_grad_wrt_delta, _ = gaussian_kernel_value_and_grad(
        sparse_points, sparse_points, delta=delta
    )
    #print("Kmm_grad_wrt_delta: ")
    #print(Kmm_grad_wrt_delta)

    # -- compute Kmn
    #kernels = gaussian_kernel(distances, sparse_points.T)
    kernels, kernel_derivs, kernel_grads_wrt_delta, _ = gaussian_kernel_value_and_grad(
        distances, sparse_points, delta=delta
    )
    #print("b2_kernels: ")
    #print(kernels)

    #print("kernel derivatives: ") # shape (num_distances, num_sparse)
    #print(kernel_derivs)

    # --- derivatives
    #print("distance derivatives: ") # shape (num_distances, 6)
    #print(dis_derivs)
        
    #print("kernels: ")
    #print(kernels)

    #print("kernel derivatives to cartesian: ") # shape (num_distances, 6, num_sparse)
    kernel_derivs = dis_derivs[:, :, np.newaxis].repeat(num_sparse, axis=2) * kernel_derivs[:, np.newaxis, :]
    #print(kernel_derivs[0])
    #print(kernel_derivs.shape)
    #print(kernel_derivs[0])

    #print(kernals)
    # --- group descriptors
    #print("===== group descriptor gradients =====")
    Knm_ene = np.zeros((nframes, num_sparse))
    Knm_ene_grad_wrt_delta = np.zeros((nframes, num_sparse))
    for i in range(nframes):
        for dis_loc, kernel, k_grad in zip(distance_mapping_list, kernels, kernel_grads_wrt_delta):
            if dis_loc[0] == i:
                Knm_ene[i, :] += kernel
                Knm_ene_grad_wrt_delta[i, :] += k_grad

    Knm_grad = np.zeros((natoms_tot*3, num_sparse))
    Knm_grad_grad_wrt_delta = np.zeros((natoms_tot*3, num_sparse))
    for dis_loc, kernel_grad in zip(distance_mapping_list, kernel_derivs):
        i = np.sum(natoms_list[:dis_loc[0]], dtype=int)*3 + dis_loc[1]*3
        j = np.sum(natoms_list[:dis_loc[0]], dtype=int)*3 + dis_loc[2]*3
        #print(i, j)
        Knm_grad[i:i+3, :] += kernel_grad[0:3, :]
        Knm_grad[j:j+3, :] += kernel_grad[3:6, :]
        Knm_grad_grad_wrt_delta[i:i+3, :] += kernel_grad[0:3, :]*2/delta
        Knm_grad_grad_wrt_delta[j:j+3, :] += kernel_grad[3:6, :]*2/delta
    Knm_frc = -Knm_grad

    #print("kernel on forces: ")
    #print(Knm_frc)

    # combine energy and force kernel
    Knm = np.vstack([Knm_ene, Knm_frc])

    #print("Knm: ")
    #print(Knm)

    return Kmm, Kmm_grad_wrt_delta, Knm, Knm_ene_grad_wrt_delta, Knm_grad_grad_wrt_delta

def compute_distance_marginal_likelihood(
    delta, sigma2, nframes, natoms_list, y_data, 
    distances, dis_derivs, distance_mapping_list, sparse_points
):
    num_points = y_data.shape[0]

    Kmm, Kmm_grad_wrt_delta, Knm, Knm_ene_grad_wrt_delta, Knm_grad_grad_wrt_delta = compute_distance_kernal_matrices(
        delta, sigma2, nframes, natoms_list, y_data, 
        distances, dis_derivs, distance_mapping_list, sparse_points
    )

    Kmm_inv = np.linalg.inv(Kmm)
    Knn = np.dot(Knm, np.dot(Kmm_inv, Knm.T)) + sigma2
    Knn_inv = np.linalg.inv(Knn)
    loss = -0.5*np.log(np.linalg.det(Knn)) - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)

    # -- get gradients
    Knm_grad_wrt_delta = np.vstack([Knm_ene_grad_wrt_delta, -Knm_grad_grad_wrt_delta])
    Kmm_inv_grad_wrt_delta = -Kmm_inv @ Kmm_grad_wrt_delta @ Kmm_inv
    Ky = Knn_inv @ y_data

    Knn_grad_wrt_delta = (
        (Knm_grad_wrt_delta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_delta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_grad_wrt_delta.T
    )

    loss_grad = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_delta)

    return -loss, -loss_grad

# --- BODY-3 ---

def gaussian_kernel_value_and_grad_body3(x1, x2, delta=0.2, theta=0.5):
    """
        x1: shape (num_points, num_feature_dim)
        x2: shape (num_sparse_points, num_feature_dim)

        v: shape (num_points, num_sparse_points)
        v: shape (num_points, num_sparse_points)
        ### g: shape (num_points, num_sparse_points, num_feature_dim)

    """
    # shape (num_points, num_sparse, num_feature_dim)
    x_diff = x1[:, np.newaxis, :].repeat(x2.shape[0], axis=1) - x2
    x_norm = np.linalg.norm(x_diff, axis=-1)

    # shape (num_points, num_sparse)
    v = delta**2*np.exp(-x_norm**2/2./theta**2)

    # gradient wrt x1
    num_feature_dim = x_diff.shape[-1]
    rv = np.repeat(v[:, :, np.newaxis], num_feature_dim, axis=-1)
    g0 = rv*(-x_diff/theta**2)

    # grad wrt delta
    g2 = v*2./delta

    # grad wrt theta
    g3 = v*(x_norm**2/theta**3)

    return (v, g0, g2, g3)


def compute_body3_kernel_matrices(
    delta, theta, r_cut, nframes, natoms_tot, 
    b3_features, body3_gradients, body3_mapping, sparse_body3_features
):
    """"""
    num_sparse = sparse_body3_features.shape[0]

    # NOTE: cutoff switch function on sparse_points
    #sparse_b3s, _ = switch_function_value_and_grad_b3(sparse_body3_features, r_cut)

    # -- consider permutations...
    num_b3 = b3_features.shape[0]
    b3_kernels = np.zeros((b3_features.shape[0], num_sparse))
    b3_kernel_grad_x1 = np.zeros((b3_features.shape[0], num_sparse, 3))
    b3ks_gdelta = np.zeros((num_b3, num_sparse))
    b3ks_gtheta = np.zeros((num_b3, num_sparse))
    #b3ks_gx1_gtheta_coef = np.zeros((num_b3, num_sparse))
    b3ks_gx1_gtheta = np.zeros((num_b3, num_sparse, 3))

    b3s, b3sg = switch_function_value_and_grad_b3(b3_features, r_cut)
    #b3s = b3s*sparse_b3s.T # shape (num_b3, num_sparse)
    #b3sg = b3sg[:, np.newaxis, :]*sparse_b3s.T[:, :, np.newaxis] # shape (num_b3, 1, 3)
    b3s = b3s # shape (num_b3, 1)
    b3sg = b3sg[:, np.newaxis, :] # shape (num_b3, 1, 3)

    for p in itertools.permutations(range(3), 3):
    #for p in [(0, 1, 2)]:
        _k, _k_g_x1, k_gdelta, k_gtheta = gaussian_kernel_value_and_grad_body3(
            b3_features, sparse_body3_features[:, p], delta=delta, theta=theta
        )
        # --
        b3_kernels += _k*b3s
        b3_kernel_grad_x1 += _k_g_x1*b3s[:, :, np.newaxis] + np.repeat(_k[:, :, np.newaxis], 3, axis=2)*b3sg
        b3ks_gdelta += k_gdelta*b3s
        b3ks_gtheta += k_gtheta*b3s
        # --
        x_norm = np.linalg.norm(
            b3_features[:, np.newaxis, :].repeat(num_sparse, axis=1)-sparse_body3_features[:, p], 
            axis=-1
        )[:, :, np.newaxis]
        #b3ks_gx1_gtheta_coef += (-2./theta + x_norm**2/theta**3)
        b3ks_gx1_gtheta += (-2./theta*_k_g_x1*b3s[:, :, np.newaxis] + x_norm**2/theta**3*b3_kernel_grad_x1)

    # --- group descriptors
    Knm_ene = np.zeros((nframes, num_sparse))
    Knm_ene_gdelta = np.zeros((nframes, num_sparse))
    Knm_ene_gtheta = np.zeros((nframes, num_sparse))
    for loc, kernel, k_gdelta, k_gtheta in zip(body3_mapping, b3_kernels, b3ks_gdelta, b3ks_gtheta):
        fi, i, j, k = loc
        Knm_ene[fi, :] += kernel          # PASS
        Knm_ene_gdelta[fi, :] += k_gdelta # PASS
        Knm_ene_gtheta[fi, :] += k_gtheta # PASS

    Knm_grad = np.zeros((natoms_tot*3, num_sparse))
    Knm_gcart_gdelta = np.zeros((natoms_tot*3, num_sparse))
    Knm_gcart_gtheta = np.zeros((natoms_tot*3, num_sparse))
    for loc, b3_grad, k_grad, k_gx1gtheta in zip(
        body3_mapping, body3_gradients, b3_kernel_grad_x1, b3ks_gx1_gtheta
    ):
        fi, i, j, k = loc
        # b3_grad (3, 6) k_grad (num_sparse, 3) -> (num_sparse, 3, 6)
        curr_grad = np.repeat(k_grad[:, :, np.newaxis], 6, axis=2) * b3_grad
        Knm_grad[i*3:i*3+3, :] += curr_grad[:, 0, 0:3].T
        Knm_grad[j*3:j*3+3, :] += curr_grad[:, 0, 3:6].T
        Knm_grad[i*3:i*3+3, :] += curr_grad[:, 1, 0:3].T
        Knm_grad[k*3:k*3+3, :] += curr_grad[:, 1, 3:6].T
        Knm_grad[j*3:j*3+3, :] += curr_grad[:, 2, 0:3].T
        Knm_grad[k*3:k*3+3, :] += curr_grad[:, 2, 3:6].T
        # -- PASS
        Knm_gcart_gdelta[i*3:i*3+3, :] += curr_grad[:, 0, 0:3].T*2./delta
        Knm_gcart_gdelta[j*3:j*3+3, :] += curr_grad[:, 0, 3:6].T*2./delta
        Knm_gcart_gdelta[i*3:i*3+3, :] += curr_grad[:, 1, 0:3].T*2./delta
        Knm_gcart_gdelta[k*3:k*3+3, :] += curr_grad[:, 1, 3:6].T*2./delta
        Knm_gcart_gdelta[j*3:j*3+3, :] += curr_grad[:, 2, 0:3].T*2./delta
        Knm_gcart_gdelta[k*3:k*3+3, :] += curr_grad[:, 2, 3:6].T*2./delta
        # -- 
        #Knm_gcart_gtheta[i*3:i*3+3, :] += (curr_grad[:, 0, 0:3]*k_gx1gtheta).T
        #Knm_gcart_gtheta[j*3:j*3+3, :] += (curr_grad[:, 0, 3:6]*k_gx1gtheta).T
        #Knm_gcart_gtheta[i*3:i*3+3, :] += (curr_grad[:, 1, 0:3]*k_gx1gtheta).T
        #Knm_gcart_gtheta[k*3:k*3+3, :] += (curr_grad[:, 1, 3:6]*k_gx1gtheta).T
        #Knm_gcart_gtheta[j*3:j*3+3, :] += (curr_grad[:, 2, 0:3]*k_gx1gtheta).T
        #Knm_gcart_gtheta[k*3:k*3+3, :] += (curr_grad[:, 2, 3:6]*k_gx1gtheta).T
        # -- b3_grad (3, 6) k_gx1gtheta (5,3) -> (3, 5, 6)
        curr_grad_theta = b3_grad[:, np.newaxis, :]*k_gx1gtheta.T[:, :, np.newaxis]
        Knm_gcart_gtheta[i*3:i*3+3, :] += curr_grad_theta[0, :, 0:3].T
        Knm_gcart_gtheta[j*3:j*3+3, :] += curr_grad_theta[0, :, 3:6].T
        Knm_gcart_gtheta[i*3:i*3+3, :] += curr_grad_theta[1, :, 0:3].T
        Knm_gcart_gtheta[k*3:k*3+3, :] += curr_grad_theta[1, :, 3:6].T
        Knm_gcart_gtheta[j*3:j*3+3, :] += curr_grad_theta[2, :, 0:3].T
        Knm_gcart_gtheta[k*3:k*3+3, :] += curr_grad_theta[2, :, 3:6].T
    Knm_frc = -Knm_grad

    Knm = np.vstack([Knm_ene, Knm_frc])
    #print("Knm shape: ")
    #print(Knm)
    #print(Knm.shape)
    #print(Knm_ene_gdelta)
    #print(Knm_gcart_gdelta)
    #print(Knm_gcart_gdelta.shape)
    #print(Knm_ene_gtheta)
    #print(Knm_gcart_gtheta)
    #print(Knm_gcart_gtheta.shape)

    # --- construct Kmm
    Kmm = np.zeros((num_sparse, num_sparse))
    Kmm_gdelta = np.zeros((num_sparse, num_sparse))
    Kmm_gtheta = np.zeros((num_sparse, num_sparse))
    # TODO: permutations?
    for p in itertools.permutations(range(3), 3):
    #for p in [(0, 1, 2)]:
        curr_Kmm, _, curr_Kmm_gdelta, curr_Kmm_gtheta = gaussian_kernel_value_and_grad_body3(
            sparse_body3_features, sparse_body3_features[:, p], delta=delta, theta=theta
        )
        Kmm += curr_Kmm
        Kmm_gdelta += curr_Kmm_gdelta
        Kmm_gtheta += curr_Kmm_gtheta

    # TODO: add cutoff function?
    #cutoff_sparse = sparse_b3s*sparse_b3s.T
    #Kmm = Kmm*cutoff_sparse

    #print("Kmm shape: ")
    #print(Kmm.shape)
    #print(Kmm)
    #exit()

    return Kmm, Knm, Kmm_gdelta, Kmm_gtheta, Knm_ene_gdelta, Knm_ene_gtheta, Knm_gcart_gdelta, Knm_gcart_gtheta

def compute_body2_kernel_matrices(
    delta, theta, r_cut, nframes, natoms_tot, 
    body2_features, body2_gradients, body2_mapping, sparse_body2_features
):
    """"""
    #print(f"r_cut: {r_cut}")
    num_sparse = sparse_body2_features.shape[0]

    #sparse_b2, _ = switch_function_value_and_grad(sparse_body2_features, r_cut)

    # - 
    _b2_kernels, _b2_kernel_grad_x1, b2_kg_wrt_delta, b2_kg_wrt_theta = gaussian_kernel_value_and_grad(
        body2_features, sparse_body2_features, delta=delta, theta=theta
    )
    #print(_b2_kernels.shape)
    #print(_b2_kernel_grad_x1.shape)
    b2s, b2sg = switch_function_value_and_grad(body2_features, r_cut)
    #b2s = b2s*sparse_b2.T # (num_b2, num_sparse)
    #b2sg = b2sg*sparse_b2.T # (num_b2, num_sparse)
    b2s = b2s # (num_b2, num_sparse)
    b2sg = b2sg # (num_b2, num_sparse)

    b2_kernels = _b2_kernels*b2s # (num_b2, num_sparse)
    b2_kernel_grad_x1 = (_b2_kernel_grad_x1*b2s+_b2_kernels*b2sg)[:, :, np.newaxis] # (num_b2, num_sparse, 1)
    #print(b2_kernel_grad_x1)
    #print(b2_kernel_grad_x1.shape)

    b2_kg_wrt_delta = (b2_kg_wrt_delta*b2s)[:, :, np.newaxis] # (num_b2, num_sparse, 1)
    b2_kg_wrt_theta = (b2_kg_wrt_theta*b2s)[:, :, np.newaxis] # (num_b2, num_sparse, 1)
    print(_b2_kernels)
    print(_b2_kernel_grad_x1)
    print(body2_gradients)

    # --- group descriptors
    #print("CONSTRUCT B2 KNM: ")
    Knm_ene = np.zeros((nframes, num_sparse))
    Knm_ene_gdelta = np.zeros((nframes, num_sparse))
    Knm_ene_gtheta = np.zeros((nframes, num_sparse))
    for loc, kernel, k_gdelta, k_gtheta in zip(body2_mapping, b2_kernels, b2_kg_wrt_delta, b2_kg_wrt_theta):
        fi, i, j = loc
        Knm_ene[fi, :] += kernel
        Knm_ene_gdelta[fi, :] += k_gdelta.squeeze() # PASS
        Knm_ene_gtheta[fi, :] += k_gtheta.squeeze() # PASS

    Knm_grad = np.zeros((natoms_tot*3, num_sparse))
    Knm_gcart_gdelta = np.zeros((natoms_tot*3, num_sparse))
    Knm_gcart_gtheta = np.zeros((natoms_tot*3, num_sparse))
    #b2ks_gx1_gtheta_coef = (-2./theta + (body2_features-sparse_body2_features.T)**2/theta**3)[:, :, np.newaxis]
    x_norm = body2_features-sparse_body2_features.T
    b2ks_gx1_gtheta = (-2./theta*_b2_kernel_grad_x1*b2s+x_norm**2/theta**3*b2_kernel_grad_x1.squeeze())[:, :, np.newaxis]

    for loc, b2_grad, k_grad, b2k_gx1gtheta in zip(
        body2_mapping, body2_gradients, b2_kernel_grad_x1, b2ks_gx1_gtheta
    ):
        fi, i, j = loc
        # b2_grad (1, 6) k_grad (num_sparse, 1) -> (num_sparse, 1, 6)
        curr_grad = np.repeat(k_grad[:, :, np.newaxis], 6, axis=2) * b2_grad
        print(f"curr_grad: ")
        print(curr_grad.squeeze())
        print("miaow: ")
        print(curr_grad[:, 0, 0:3].T)
        Knm_grad[i*3:i*3+3, :] += curr_grad[:, 0, 0:3].T
        print(curr_grad[:, 0, 3:6].T)
        Knm_grad[j*3:j*3+3, :] += curr_grad[:, 0, 3:6].T
        print(Knm_grad)
        # --
        Knm_gcart_gdelta[i*3:i*3+3, :] += curr_grad[:, 0, 0:3].T*2./delta
        Knm_gcart_gdelta[j*3:j*3+3, :] += curr_grad[:, 0, 3:6].T*2./delta
        # --
        #Knm_gcart_gtheta[i*3:i*3+3, :] += (curr_grad[:, 0, 0:3]*b2k_gx1gtheta).T
        #Knm_gcart_gtheta[j*3:j*3+3, :] += (curr_grad[:, 0, 3:6]*b2k_gx1gtheta).T
        #xxx = b2_grad[:, 0:3]*b2k_gx1gtheta
        #print(xxx.shape)
        #exit()
        Knm_gcart_gtheta[i*3:i*3+3, :] += (b2_grad[:, 0:3]*b2k_gx1gtheta).T
        Knm_gcart_gtheta[j*3:j*3+3, :] += (b2_grad[:, 3:6]*b2k_gx1gtheta).T
    Knm_frc = -Knm_grad

    Knm = np.vstack([Knm_ene, Knm_frc])
    print("Knm: ")
    print(Knm)
    #print(Knm.shape)
    #print(Knm_ene_gdelta)
    #print(Knm_gcart_gdelta)
    #print(Knm_gcart_gdelta.shape)
    #print(Knm_ene_gtheta)
    #print(Knm_gcart_gtheta)
    #print(Knm_gcart_gtheta.shape)
    # ---

    # --- construct Kmm
    Kmm, _, Kmm_gdelta, Kmm_gtheta  = gaussian_kernel_value_and_grad(
        sparse_body2_features, sparse_body2_features, delta=delta, theta=theta
    )
    # NOTE: add cutoff function?
    #cutoff_sparse = sparse_b2*sparse_b2.T
    #Kmm = Kmm*cutoff_sparse
    #Kmm_gdelta = Kmm_gdelta*cutoff_sparse
    #Kmm_gtheta = Kmm_gtheta*cutoff_sparse

    #print("Kmm shape: ", Kmm.shape)

    return Kmm, Knm, Kmm_gdelta, Kmm_gtheta, Knm_ene_gdelta, Knm_ene_gtheta, Knm_gcart_gdelta, Knm_gcart_gtheta

def compute_b2_marginal_likelihood(
    params, sigma2, r_cut, y_data, nframes, natoms_tot, 
    body2_features, body2_gradients, body2_mapping, sparse_body2_features
):
    delta, theta = params

    num_points = y_data.shape[0]

    Kmm, Knm, Kmm_gdelta, Kmm_gtheta, Knm_ene_gdelta, Knm_ene_gtheta, Knm_gcart_gdelta, Knm_gcart_gtheta = compute_body2_kernel_matrices(
        delta, theta, r_cut, nframes, natoms_tot, 
        body2_features, body2_gradients, body2_mapping, sparse_body2_features
    )
    Kmm_inv = np.linalg.inv(Kmm)

    #print("Knm@Kmn: ")
    #print(np.linalg.det(Knm@Knm.T))
    #print(np.linalg.det(Knm[:nframes, :]@Knm[:nframes, :].T))

    Knn = np.dot(Knm, np.dot(Kmm_inv, Knm.T)) + sigma2
    Knn_inv = np.linalg.inv(Knn)
    print("Knn: ")
    print(Knn)
    np.savetxt("./xxx.dat", Knn, fmt="%12.4f")
    print(np.linalg.det(Knn))
    #print(Knn.shape)
    exit()

    loss = -0.5*np.log(np.linalg.det(Knn)) - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)

    # -- get gradients
    Knm_grad_wrt_delta = np.vstack([Knm_ene_gdelta, -Knm_gcart_gdelta])
    Kmm_inv_grad_wrt_delta = -Kmm_inv @ Kmm_gdelta @ Kmm_inv

    Knn_grad_wrt_delta = (
        (Knm_grad_wrt_delta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_delta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_grad_wrt_delta.T
    )

    Knm_grad_wrt_theta = np.vstack([Knm_ene_gtheta, -Knm_gcart_gtheta])
    Kmm_inv_grad_wrt_theta = -Kmm_inv @ Kmm_gtheta @ Kmm_inv

    Knn_grad_wrt_theta = (
        (Knm_grad_wrt_theta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_theta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_grad_wrt_theta.T
    )

    Ky = Knn_inv @ y_data

    loss_gdelta = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_delta)
    loss_gtheta = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_theta)

    return -loss[0][0], [-loss_gdelta, -loss_gtheta]

def compute_b3_marginal_likelihood(
    params, sigma2, r_cut, y_data, nframes, natoms_tot, 
    b3_features, b3_gradients, b3_mapping, sparse_b3_features
):
    delta, theta = params

    num_points = y_data.shape[0]

    Kmm, Knm, Kmm_gdelta, Kmm_gtheta, Knm_ene_gdelta, Knm_ene_gtheta, Knm_gcart_gdelta, Knm_gcart_gtheta = compute_body3_kernel_matrices(
        delta, theta, r_cut, nframes, natoms_tot, 
        b3_features, b3_gradients, b3_mapping, sparse_b3_features
    )

    Kmm_inv = np.linalg.inv(Kmm)
    Knn = np.dot(Knm, np.dot(Kmm_inv, Knm.T)) + sigma2
    Knn_inv = np.linalg.inv(Knn)

    print("Knn: ")
    print(Knn)
    print(np.linalg.det(Knn))

    loss = -0.5*np.log(np.linalg.det(Knn)) - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)

    # -- get gradients
    Knm_grad_wrt_delta = np.vstack([Knm_ene_gdelta, -Knm_gcart_gdelta])
    Kmm_inv_grad_wrt_delta = -Kmm_inv @ Kmm_gdelta @ Kmm_inv

    Knn_grad_wrt_delta = (
        (Knm_grad_wrt_delta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_delta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_grad_wrt_delta.T
    )

    Knm_grad_wrt_theta = np.vstack([Knm_ene_gtheta, -Knm_gcart_gtheta])
    Kmm_inv_grad_wrt_theta = -Kmm_inv @ Kmm_gtheta @ Kmm_inv

    Knn_grad_wrt_theta = (
        (Knm_grad_wrt_theta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_theta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_grad_wrt_theta.T
    )

    Ky = Knn_inv @ y_data

    loss_gdelta = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_delta)
    loss_gtheta = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_theta)

    # NOTE: LOSS should be a scalar otherwise bug?
    return -loss[0][0], [-loss_gdelta, -loss_gtheta]

def compute_b2b3_marginal_likelihood(
    params, sigma2, r_cut, y_data, nframes, natoms_tot, 
    body2_features, body2_gradients, body2_mapping, sparse_b2_features, 
    body3_features, body3_gradients, body3_mapping, sparse_b3_features
):
    # -
    b2_delta, b2_theta, b3_delta, b3_theta = params

    # --
    num_points = y_data.shape[0]

    # - body2
    (
        b2_Kmm, b2_Knm, 
        b2_Kmm_gdelta, b2_Kmm_gtheta, 
        b2_Knm_ene_gdelta, b2_Knm_ene_gtheta, 
        b2_Knm_gcart_gdelta, b2_Knm_gcart_gtheta
    ) = compute_body2_kernel_matrices(
        b2_delta, b2_theta, r_cut, nframes, natoms_tot, 
        body2_features, body2_gradients, body2_mapping, sparse_b2_features
    )
    b2_Knm_gdelta = np.vstack([b2_Knm_ene_gdelta, -b2_Knm_gcart_gdelta])
    b2_Knm_gtheta = np.vstack([b2_Knm_ene_gtheta, -b2_Knm_gcart_gtheta])

    # -
    (
        b3_Kmm, b3_Knm, 
        b3_Kmm_gdelta, b3_Kmm_gtheta, 
        b3_Knm_ene_gdelta, b3_Knm_ene_gtheta, 
        b3_Knm_gcart_gdelta, b3_Knm_gcart_gtheta
    ) = compute_body3_kernel_matrices(
        b3_delta, b3_theta, r_cut, nframes, natoms_tot, 
        body3_features, body3_gradients, body3_mapping, sparse_b3_features
    )
    b3_Knm_gdelta = np.vstack([b3_Knm_ene_gdelta, -b3_Knm_gcart_gdelta])
    b3_Knm_gtheta = np.vstack([b3_Knm_ene_gtheta, -b3_Knm_gcart_gtheta])

    num_b2_sparse = sparse_b2_features.shape[0]
    num_b3_sparse = sparse_b3_features.shape[0]

    #loss, lg_b2delta, lg_b2theta, lg_b3delta, lg_b3theta = train_ene_and_frc(
    #    nframes, num_points, num_b2_sparse, num_b3_sparse, y_data, sigma2, b2_Kmm, b3_Kmm, b2_Knm, b3_Knm,
    #    b2_Kmm_gdelta, b2_Kmm_gtheta, b2_Knm_gdelta, b2_Knm_gtheta,
    #    b3_Kmm_gdelta, b3_Kmm_gtheta, b3_Knm_gdelta, b3_Knm_gtheta,
    #)
    loss, lg_b2delta, lg_b2theta, lg_b3delta, lg_b3theta = train_ene(
        nframes, num_points, num_b2_sparse, num_b3_sparse, y_data, sigma2, b2_Kmm, b3_Kmm, b2_Knm, b3_Knm,
        b2_Kmm_gdelta, b2_Kmm_gtheta, b2_Knm_gdelta, b2_Knm_gtheta,
        b3_Kmm_gdelta, b3_Kmm_gtheta, b3_Knm_gdelta, b3_Knm_gtheta,
    )

    return -loss[0][0], [-lg_b2delta, -lg_b2theta, -lg_b3delta, -lg_b3theta]

def train_ene(
    nframes, num_points, num_b2_sparse, num_b3_sparse, y_data, sigma2, b2_Kmm, b3_Kmm, b2_Knm, b3_Knm,
    b2_Kmm_gdelta, b2_Kmm_gtheta, b2_Knm_gdelta, b2_Knm_gtheta,
    b3_Kmm_gdelta, b3_Kmm_gtheta, b3_Knm_gdelta, b3_Knm_gtheta,
):
    # - train ene
    Kmm = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm[:num_b2_sparse, :num_b2_sparse] = b2_Kmm
    Kmm[num_b2_sparse:, num_b2_sparse:] = b3_Kmm

    Knm = np.zeros((nframes, num_b2_sparse+num_b3_sparse))
    Knm[:, :num_b2_sparse] = b2_Knm[:nframes, :]
    Knm[:, num_b2_sparse:] = b3_Knm[:nframes, :]

    Kmm_inv = np.linalg.inv(Kmm)
    Knn = Knm@Kmm_inv@Knm.T + sigma2[:nframes, :nframes]
    print("Knn: ")
    print(np.linalg.det(Knn))

    Knn_inv = np.linalg.inv(Knn)

    y_data_ene = y_data[:nframes, :]

    sign, logabsdet = np.linalg.slogdet(Knn)
    loss = -0.5*(np.log(sign)+logabsdet) - 0.5 * y_data_ene.T @ Knn_inv @ y_data_ene - nframes/2.*np.log(2*np.pi)

    Ky = Knn_inv @ y_data_ene

    Kmm_gdelta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gdelta[:num_b2_sparse, :num_b2_sparse] = b2_Kmm_gdelta
    Knm_gdelta = np.zeros((nframes, num_b2_sparse+num_b3_sparse))
    Knm_gdelta[:, :num_b2_sparse] = b2_Knm_gdelta[:nframes, :]
    lg_b2delta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gdelta, Knm, Knm_gdelta)

    Kmm_gtheta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gtheta[:num_b2_sparse, :num_b2_sparse] = b2_Kmm_gtheta
    Knm_gtheta = np.zeros((nframes, num_b2_sparse+num_b3_sparse))
    Knm_gtheta[:, :num_b2_sparse] = b2_Knm_gtheta[:nframes, :]
    lg_b2theta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gtheta, Knm, Knm_gtheta)

    Kmm_gdelta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gdelta[num_b2_sparse:, num_b2_sparse:] = b3_Kmm_gdelta
    Knm_gdelta = np.zeros((nframes, num_b2_sparse+num_b3_sparse))
    Knm_gdelta[:, num_b2_sparse:] = b3_Knm_gdelta[:nframes, :]
    lg_b3delta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gdelta, Knm, Knm_gdelta)

    Kmm_gtheta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gtheta[num_b2_sparse:, num_b2_sparse:] = b3_Kmm_gtheta
    Knm_gtheta = np.zeros((nframes, num_b2_sparse+num_b3_sparse))
    Knm_gtheta[:, num_b2_sparse:] = b3_Knm_gtheta[:nframes, :]
    lg_b3theta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gtheta, Knm, Knm_gtheta)

    return loss, lg_b2delta, lg_b2theta, lg_b3delta, lg_b3theta

def train_ene_and_frc(
    nframes, num_points, num_b2_sparse, num_b3_sparse, y_data, sigma2, b2_Kmm, b3_Kmm, b2_Knm, b3_Knm,
    b2_Kmm_gdelta, b2_Kmm_gtheta, b2_Knm_gdelta, b2_Knm_gtheta,
    b3_Kmm_gdelta, b3_Kmm_gtheta, b3_Knm_gdelta, b3_Knm_gtheta,
):
    """"""
    Kmm = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm[:num_b2_sparse, :num_b2_sparse] = b2_Kmm
    Kmm[num_b2_sparse:, num_b2_sparse:] = b3_Kmm

    Knm = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
    Knm[:, :num_b2_sparse] = b2_Knm
    Knm[:, num_b2_sparse:] = b3_Knm

    # - combine
    Kmm_inv = np.linalg.inv(Kmm)
    print("Kmm_inv: ")
    print(np.linalg.det(Kmm_inv))

    # --- approximate Knn
    # -0.
    #print("Knm@Kmn: ")
    #print(np.linalg.det(Knm@Knm.T))
    #print(np.linalg.det(Knm[:nframes, :]@Knm[:nframes, :].T))
    #eigens, eig_vecs = np.linalg.eig(Kmm)
    #eigens = eigens[:, np.newaxis]
    #mu = np.sqrt((num_b2_sparse+num_b3_sparse)/num_points)*eigens**-1*Knm@eig_vecs
    #Knn = (num_points/(num_b2_sparse+num_b3_sparse))*eigens*mu@mu.T
    #Knn += sigma2
    #Knn_inv = np.linalg.inv(Knn)

    # -1.
    Knn = Knm@Kmm_inv@Knm.T + sigma2
    Knn_inv = np.linalg.inv(Knn)

    # -2.
    #Knn = Kmm
    #Knn_inv = Kmm_inv

    print("Knn: ")
    #print(Knn)
    print(np.linalg.det(Knn))
    #print(np.log(np.linalg.det(Knn)))
    #print(np.linalg.slogdet(Knn))
    sign, logabsdet = np.linalg.slogdet(Knn)
    #print(sign * np.exp(logabsdet))
    #exit()
    # --- approximate Knn

    # NOTE: det(A) will be zero if there are rows with zeros since
    #       some atoms may not in any three-body...
    #loss = -0.5*np.log(np.linalg.det(Knn)) - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)
    loss = -0.5*(np.log(sign)+logabsdet) - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)
    #loss = - 0.5 * y_data.T @ Knn_inv @ y_data - num_points/2.*np.log(2*np.pi)

    # -- combine gradients
    Ky = Knn_inv @ y_data

    Kmm_gdelta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gdelta[:num_b2_sparse, :num_b2_sparse] = b2_Kmm_gdelta
    Knm_gdelta = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
    Knm_gdelta[:, :num_b2_sparse] = b2_Knm_gdelta
    lg_b2delta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gdelta, Knm, Knm_gdelta)

    Kmm_gtheta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gtheta[:num_b2_sparse, :num_b2_sparse] = b2_Kmm_gtheta
    Knm_gtheta = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
    Knm_gtheta[:, :num_b2_sparse] = b2_Knm_gtheta
    lg_b2theta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gtheta, Knm, Knm_gtheta)

    Kmm_gdelta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gdelta[num_b2_sparse:, num_b2_sparse:] = b3_Kmm_gdelta
    Knm_gdelta = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
    Knm_gdelta[:, num_b2_sparse:] = b3_Knm_gdelta
    lg_b3delta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gdelta, Knm, Knm_gdelta)

    Kmm_gtheta = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
    Kmm_gtheta[num_b2_sparse:, num_b2_sparse:] = b3_Kmm_gtheta
    Knm_gtheta = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
    Knm_gtheta[:, num_b2_sparse:] = b3_Knm_gtheta
    lg_b3theta = compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gtheta, Knm, Knm_gtheta)

    return loss, lg_b2delta, lg_b2theta, lg_b3delta, lg_b3theta

def compute_loss_gradient(Ky, Knn_inv, Kmm_inv, Kmm_gdelta, Knm, Knm_gdelta):
    """"""
    Kmm_inv_grad_wrt_delta = -Kmm_inv @ Kmm_gdelta @ Kmm_inv
    Knn_grad_wrt_delta = (
        (Knm_gdelta @ Kmm_inv + Knm @ Kmm_inv_grad_wrt_delta) @ Knm.T +
        Knm @ Kmm_inv @ Knm_gdelta.T
    )

    loss_gdelta = 0.5*np.trace((Ky @ Ky.T - Knn_inv) @ Knn_grad_wrt_delta)

    return loss_gdelta


class SparseGaussianProcessTrainer():

    def __init__(self) -> None:
        """"""
        self.r_cut = 4.0
        #self.r_cut = 6.8
        #self.r_cut = 8.0

        return
    
    def _prepare_dataset(self, ):
        """"""
        # - read dataset
        #frames = read("./dft/O2.xyz", ":1")
        #frames = read("./dft/Cu4.xyz", ":1")
        frames = read("./dft/Cu4.xyz", "1:2")
        #frames = read("./dft/Cu13.xyz", ":")
        #frames = read("./md/Cu4/cand0/traj.xyz", ":")
        #frames = read("./Cu100/cand0/traj.xyz", ":5")

        energies = [a.get_potential_energy() for a in frames]
        energies = np.array(energies)[:, np.newaxis]

        forces = np.vstack([a.get_forces() for a in frames])
        forces = forces.flatten()[:, np.newaxis]
        print(f"force shape: {forces.shape}")

        y_data = np.vstack([energies, forces])
        print(y_data)

        # - regularization
        sigma2_diag_ene = 0.05*np.sqrt(np.array([len(a) for a in frames]))
        f_sig, f_min = 0.01, 0.01
        sigma2_diag_frc = []
        for x in np.fabs(forces.flatten()):
            if x >= f_min:
                sigma2_diag_frc.append(f_sig*x)
            else:
                sigma2_diag_frc.append(f_sig*f_min)
        sigma2_diag = sigma2_diag_ene.tolist() + sigma2_diag_frc
        sigma2 = np.diag(sigma2_diag)**2

        return frames, y_data, sigma2
    
    def run(self, *args, **kwargs):
        """"""
        # - read dataset
        frames, y_data, sigma2 = self._prepare_dataset()
        nframes = len(frames)
        natoms_list = [len(a) for a in frames]
        natoms_tot = np.sum(natoms_list)

        num_points = nframes + natoms_tot*3

        # - get atomic environments
        (
            body2_mapping, body2_features, body2_gradients, body3_mapping, body3_features, body3_gradients
        ) = compute_body3_descriptor(frames, self.r_cut)
        #print("BODY3 MAPPING: ")
        #print(body3_mapping)

        #print("BODY3 GRADIENTS: ")
        #print(body3_gradients)

        # - sparsify
        sparse_body2_features = np.loadtxt("./b2_sparse.dat")[:, np.newaxis]

        print("BODY3 SHAPE: ")
        print(body3_features.shape)
        from gdpx.selector.cur import cur_selection, fps_selection
        #selected_indices = range(0, 30, 3)
        #_, selected_indices = cur_selection(body3_features, num=20, zeta=-1)
        #_, selected_indices = fps_selection(body3_features, num=10, min_distance=0.5)
        #sparse_body3_features = body3_features[selected_indices, :]
        sparse_body3_features = np.loadtxt("./backup/b3_sparse.dat")
        print(sparse_body3_features)

        # - construct matrix
        # -- parameters
        delta, theta = 0.1, 0.5

        # --- body2
        print("~~~~~ USE BODY-2 ~~~~~")
        Kmm_b2, Knm_b2, _, _, _, _, _, _ = compute_body2_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features
        )

        # --- body3
        print("~~~~~ USE BODY-3 ~~~~~")
        Kmm_b3, Knm_b3, _, _, _, _, _, _ = compute_body3_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        )

        # -- combine matrices
        num_b2_sparse = sparse_body2_features.shape[0]
        num_b3_sparse = sparse_body3_features.shape[0]

        Kmm = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
        Kmm[:num_b2_sparse, :num_b2_sparse] = Kmm_b2
        Kmm[num_b2_sparse:, num_b2_sparse:] = Kmm_b3

        Knm = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
        Knm[:, :num_b2_sparse] = Knm_b2
        Knm[:, num_b2_sparse:] = Knm_b3

        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        ret = compute_b2b3_marginal_likelihood(
            [delta, theta, delta, theta],
            sigma2, self.r_cut, y_data, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        )
        print(ret)

        # ---
        res = optimize.minimize(
            compute_b2b3_marginal_likelihood, [delta, theta, delta, theta],
            args=(
                sigma2, self.r_cut, y_data, nframes, natoms_tot, 
                body2_features, body2_gradients, body2_mapping, sparse_body2_features, 
                body3_features, body3_gradients, body3_mapping, sparse_body3_features
            ), jac=True, options={"disp": True, "maxiter": 10}
        )
        print("OPT INFO: ")
        print(res)

        # ---
        params = res.x
        Kmm_b2, Knm_b2, _, _, _, _, _, _ = compute_body2_kernel_matrices(
            params[0], params[1], self.r_cut, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features
        )
        Kmm_b3, Knm_b3, _, _, _, _, _, _ = compute_body3_kernel_matrices(
            params[2], params[3], self.r_cut, nframes, natoms_tot, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        )

        # -- combine matrices
        num_b2_sparse = sparse_body2_features.shape[0]
        num_b3_sparse = sparse_body3_features.shape[0]

        Kmm = np.zeros((num_b2_sparse+num_b3_sparse, num_b2_sparse+num_b3_sparse))
        Kmm[:num_b2_sparse, :num_b2_sparse] = Kmm_b2
        Kmm[num_b2_sparse:, num_b2_sparse:] = Kmm_b3

        Knm = np.zeros((num_points, num_b2_sparse+num_b3_sparse))
        Knm[:, :num_b2_sparse] = Knm_b2
        Knm[:, num_b2_sparse:] = Knm_b3

        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        return

    def run_b2(self, *args, **kwargs):
        """"""
        # - read dataset
        frames, y_data, sigma2 = self._prepare_dataset()
        nframes = len(frames)
        natoms_list = [len(a) for a in frames]
        natoms_tot = np.sum(natoms_list)

        # - get atomic environments
        (
            body2_mapping, body2_features, body2_gradients, body3_mapping, body3_features, body3_gradients
        ) = compute_body3_descriptor(frames, self.r_cut)
        print(body2_mapping)

        # - construct matrix
        # -- parameters
        delta, theta = 0.5, 0.8
        #delta, theta = 100, 1000

        # --- body2
        print("~~~~~ USE BODY-2 ~~~~~")
        sparse_body2_features = np.array([0.5, 1.0, 1.5, 2.0, 2.5])[:, np.newaxis]
        Kmm_b2, Knm_b2, _, _, _, _, _, _ = compute_body2_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features
        )

        # - init train
        Kmm = Kmm_b2
        Knm = Knm_b2
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        ret = compute_b2_marginal_likelihood(
            [delta, theta],
            sigma2, self.r_cut, y_data, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features
        )
        print([delta, theta])
        print(ret)

        res = optimize.minimize(
            compute_b2_marginal_likelihood, [delta, theta],
            args=(
                sigma2, self.r_cut, y_data, nframes, natoms_tot, 
                body2_features, body2_gradients, body2_mapping, sparse_body2_features
            ), jac=True, options={"disp": True}
        )
        print("OPT INFO: ")
        print(res)
        delta, theta = res.x
        Kmm_b2, Knm_b2, _, _, _, _, _, _ = compute_body2_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body2_features, body2_gradients, body2_mapping, sparse_body2_features
        )
        Kmm = Kmm_b2
        Knm = Knm_b2
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        return

    def run_b3(self, *args, **kwargs):
        """"""
        # - read dataset
        frames, y_data, sigma2 = self._prepare_dataset()
        nframes = len(frames)
        natoms_list = [len(a) for a in frames]
        natoms_tot = np.sum(natoms_list)

        num_points = nframes + natoms_tot*3

        # - get atomic environments
        (
            body2_mapping, body2_features, body2_gradients, body3_mapping_, body3_features_, body3_gradients_
        ) = compute_body3_descriptor(frames, self.r_cut)

        # - remove duplicates?
        candidates = [0, 1]
        body3_mapping = [body3_mapping_[c] for c in candidates]
        body3_features = body3_features_[candidates, :]
        body3_gradients = body3_gradients_[candidates, :]

        print("BODY3 SHAPE: ")
        print(body3_features.shape)
        #sparse_body3_features = body3_features[:100, :]

        # - construct matrix
        # -- parameters
        delta, theta = 0.5, 0.8

        # --- body3
        print("~~~~~ USE BODY-3 ~~~~~")
        sparse_body3_features = np.array(
            [
                [1, 2, 3],
                [1.5, 2.5, 3.5],
                [2, 3, 4],
                [2.5, 3.5, 4.5],
                [4, 5, 6],
            ]
        )
        Kmm_b3, Knm_b3, _, _, _, _, _, _ = compute_body3_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        )
        # - init train
        Kmm = Kmm_b3
        Knm = Knm_b3
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        ret = compute_b3_marginal_likelihood(
            [delta, theta],
            sigma2, self.r_cut, y_data, nframes, natoms_tot, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        ) 
        print([delta, theta])
        print(ret)

        res = optimize.minimize(
            compute_b3_marginal_likelihood, [delta, theta],
            args=(
                sigma2, self.r_cut, y_data, nframes, natoms_tot, 
                body3_features, body3_gradients, body3_mapping, sparse_body3_features
            ), jac=True, options={"disp": True, "maxiter": 100}
        )
        print("OPT INFO: ")
        print(res)
        delta, theta = res.x
        Kmm_b3, Knm_b3, _, _, _, _, _, _ = compute_body3_kernel_matrices(
            delta, theta, self.r_cut, nframes, natoms_tot, 
            body3_features, body3_gradients, body3_mapping, sparse_body3_features
        )
        Kmm = Kmm_b3
        Knm = Knm_b3

        # - train
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        return
    
    def _prev_run2(self, *args, **kwargs):
        """"""
        # - read dataset
        frames, y_data, sigma2 = self._prepare_dataset()
        nframes = len(frames)
        natoms_list = [len(a) for a in frames]
        natoms_tot = np.sum(natoms_list)

        # - get atomic environments
        distances, distance_mapping_list, dis_derivs = compute_body2_descriptor(
            frames, self.r_cut
        )
        print("BODY-2 FEATURE GRADIENTS W.R.T CARTESIAN: ")
        print(dis_derivs)

        # - train
        # -- parameters
        delta = 0.2
        sparse_points = np.array([0.5, 1.0, 1.5, 2.0, 2.5])[:, np.newaxis]

        # -- compute coefficients
        Kmm, _, Knm, _, _ = compute_distance_kernal_matrices(
            delta, sigma2, nframes, natoms_list, y_data, 
            distances, dis_derivs, distance_mapping_list, sparse_points
        )
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        loss = compute_distance_marginal_likelihood(
            delta, sigma2, nframes, natoms_list, y_data, 
            distances, dis_derivs, distance_mapping_list, sparse_points
        )
        print(f"LOSS: {loss}")

        # - train
        res = optimize.minimize(
            compute_distance_marginal_likelihood, delta,
            args=(
                sigma2, nframes, natoms_list, y_data, 
                distances, dis_derivs, distance_mapping_list, sparse_points
            ), jac=True
        )
        print("OPT INFO: ")
        print(res)

        ## -- compute coefficients
        Kmm, _, Knm, _, _ = compute_distance_kernal_matrices(
            res.x, sigma2, nframes, natoms_list, y_data, 
            distances, dis_derivs, distance_mapping_list, sparse_points
        )
        self._train_and_predict(nframes, sigma2, y_data, Kmm, Knm)

        return
    
    def _train_and_predict(self, nframes, sigma2_matrix, y_data, Kmm, Knm):
        """"""
        jitter = 1e-5*np.eye(Kmm.shape[0])
        inverseLamb = np.linalg.inv(sigma2_matrix)

        # - direct inversion
        #weights = (
        #    np.linalg.inv(Kmm + jitter + Knm.T@inverseLamb@Knm) @ Knm.T @ inverseLamb
        #)
        #weights = np.dot(weights, y_data)
        #print("weights: ")
        #print(weights)

        # - decomposition
        wKmn = Knm.T@inverseLamb
        Qmm = Kmm + jitter + wKmn@Knm
        Vm = wKmn@y_data
        from scipy.linalg import solve
        weights = solve(Qmm, Vm, assume_a="pos")

        # - test
        print("===== TEST =====")
        #print("DFT: ")
        dft_data = y_data.flatten()
        #print("SGP: ")
        predictions = np.dot(Knm, weights)
        sgp_data  = predictions.flatten()
        print("ERR: ")
        ene_rmse = np.sqrt(np.sum((dft_data[:nframes] - sgp_data[:nframes])**2))
        ene_svar = np.sqrt(np.var((dft_data[:nframes] - sgp_data[:nframes])**2))
        print(f"ene_rmse: {ene_rmse} _ {ene_svar}")
        frc_rmse = np.sqrt(np.sum((dft_data[nframes:] - sgp_data[nframes:])**2))
        frc_svar = np.sqrt(np.var((dft_data[nframes:] - sgp_data[nframes:])**2))
        print(f"frc_rmse: {frc_rmse} _ {frc_svar}")

        return


class SGPTrainer(AbstractTrainer):

    name = "sgp"

    def __init__(
        self, config: dict, type_list: List[str] = None, train_epochs: int = 200, 
        directory=".", command="train", freeze_command="freeze", random_seed: int = None, 
        *args, **kwargs
    ) -> None:
        """"""
        super().__init__(
            config, type_list, train_epochs, 
            directory, command, freeze_command, 
            random_seed, *args, **kwargs
        )

        return

    @property
    def frozen_name(self):
        """"""
        return f"{self.name}.pb"

    def _resolve_freeze_command(self, *args, **kwargs):
        return super()._resolve_freeze_command(*args, **kwargs)

    def _resolve_train_command(self, *args, **kwargs):
        """"""
        command = self.command

        return command
    
    def train(self, dataset, init_model=None, *args, **kwargs):
        """"""
        self._print("miaow")
        sgp_trainer = SparseGaussianProcessTrainer()
        #print("!!!!! BODY-2 !!!!!") 
        #sgp_trainer._prev_run2()
        #print("!!!!! BODY-2 !!!!!") 
        #sgp_trainer.run_b2()
        #print("!!!!! BODY-3 !!!!!") 
        sgp_trainer.run_b3()
        #print("!!!!! BODY-2+3 !!!!!") 
        #sgp_trainer.run()

        return
    
    def write_input(self, dataset, *args, **kwargs):
        return super().write_input(dataset, *args, **kwargs)

    def read_convergence(self) -> bool:
        return super().read_convergence()


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
from typing import List

import numpy as np

import ase
from ase import Atoms
from ase.io import read, write
from ase.ga.utilities import closest_distances_generator, atoms_too_close

from .builder import StructureModifier


"""Pack given molecules into given box.

This is a modifer as it requires substrates when initialising.

"""


class PackerBuilder(StructureModifier):

    #: Number of attempts to create a random candidate.
    MAX_ATTEMPTS_PER_CANDIDATE: int = 1000

    #: Number of attempts to create a number of candidates.
    #       if 10 structures are to create, run will try 5*10=50 times.
    MAX_TIMES_SIZE: int = 10

    def __init__(
        self, substrates=None, numbers: List[int]=None, box=np.eye(3)*20., 
        covalent_ratio=[1.0, 2.0], intermoleculer_distance=[-np.inf, np.inf],
        *args, **kwargs
    ):
        """"""
        # This sets directory and random_seed.
        super().__init__(substrates=substrates, *args, **kwargs) 

        self.numbers = numbers

        # TODO: replace this with Region
        self.box = box

        # distance restraints
        self.covalent_min = covalent_ratio[0]
        self.covalent_max = covalent_ratio[1]

        self.intermol_min = intermoleculer_distance[0]
        self.intermol_max = intermoleculer_distance[1]

        return
    
    def run(self, substrates: List[Atoms]=None, size: int=1, *args, **kwargs) -> List[Atoms]:
        """Modify input structures.

        Args:
            substrates: Building blocks.
            numbers: Number of each block.
        
        Returns:
            A list of structures.

        """
        super().run(substrates=substrates, *args, **kwargs)

        n_substrates = len(self.substrates)
        if self.numbers is None:
            self.numbers = [1]*n_substrates
        n_numbers = len(self.numbers)
        assert n_numbers == n_substrates, "Incosistent number of substrates and their numbers."

        # create neighbour list
        atomic_numbers = []
        for atoms in self.substrates:
            atomic_numbers.extend(atoms.get_atomic_numbers())
        unique_atomic_numbers = list(set(atomic_numbers))
        blmin = closest_distances_generator(
            atom_numbers=unique_atomic_numbers,
            ratio_of_covalent_radii = self.covalent_min # be careful with test too far
        )
        # print(blmin)

        # - run over...
        frames = []
        for i in range(size*self.MAX_TIMES_SIZE):
            nframes = len(frames)
            if nframes < size:
                atoms = self._irun(blmin=blmin)
                if atoms is not None:
                    frames.append(atoms)
                    nframes += 1
            else:
                break
        else:
            raise RuntimeError(f"Failed to create {size} structures, only {nframes} are created.")

        return frames
    
    def _irun(self, blmin: dict, *args, **kwargs) -> Atoms:
        """"""
        box = self.box
        molecules = []
        for n, m in zip(self.numbers, self.substrates):
            molecules.extend([copy.deepcopy(m) for i in range(n)])

        # TODO: check and overwrite tags
        for i, atoms in enumerate(molecules):
            atoms.set_tags(i)

        n_molecules = len(molecules)

        packed_structure = None
        for i in range(self.MAX_ATTEMPTS_PER_CANDIDATE):
            # -- get new molecule configurations
            curr_cops = np.dot(self.rng.random((n_molecules,3)), box)
            # -- check intermolecular distances
            pairs = itertools.combinations(curr_cops, 2)
            pair_distances = np.array(
                [*itertools.starmap(lambda p1, p2: np.linalg.norm(p1-p2), pairs)]
            )
            #print(pair_distances)
            if np.any(pair_distances > self.intermol_max) or np.any(pair_distances < self.intermol_min):
                continue
            # -- translate molecules
            for cop, atoms in zip(curr_cops, molecules):
                self._translate(atoms, cop)
            # -- rotate molecules
            for j in range(self.MAX_ATTEMPTS_PER_CANDIDATE):
                packed_structure = Atoms()
                for atoms in molecules:
                    self._rotate(atoms)
                    packed_structure += atoms
                packed_structure.set_cell(box, scale_atoms=False, apply_constraint=True)
                # -- check atomic distances
                if not atoms_too_close(packed_structure, blmin, use_tags=True):
                    break
            else:
                packed_structure = None
            if packed_structure is not None:
                break
        else:
            #raise RuntimeError("Failed to pack molecules. Try increase max_attempts.")
            ...
        
        # - centre the atoms
        if packed_structure is not None:
            self._translate(packed_structure, np.sum(box/2., axis=0))

        return packed_structure
    
    def _translate(self, atoms: Atoms, dest) -> Atoms:
        """"""
        cop = np.average(atoms.get_positions(), axis=0)
        atoms.translate(dest - cop)

        return atoms
    
    def _rotate(self, atoms: Atoms) -> Atoms:
        """"""
        cop = np.average(atoms.get_positions(), axis=0)
        phi, theta, psi = 360 * self.rng.uniform(0,1,3)
        atoms.euler_rotate(
            phi=phi, theta=0.5 * theta, psi=psi,
            center=cop
        )

        return atoms


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

"""Some shared configuration parameters.
"""

import logging
from typing import Union, List, Callable

#: 
logger = logging.getLogger("GDP")
logger.setLevel(logging.INFO)

formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s: %(message)s",
    datefmt="%Y%b%d-%H:%M:%S"
    #"%(levelname)s: %(module)s - %(message)s"
)
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)

_print: Callable = logger.info
_debug: Callable = logger.debug

LOGO_LINES = [
"  ____ ____  ______  __ ",
" / ___|  _ \|  _ \ \/ / ",
"| |  _| | | | |_) \  /  ",
"| |_| | |_| |  __//  \  ",
" \____|____/|_|  /_/\_\ ",
"                        ",
]

#: Number of parallel jobs for joblib.
NJOBS: int = 1

#: Global random number generator
GRNG = None

# - find default vasp settings
#gdpconfig = Path.home() / ".gdp"
#if gdpconfig.exists() and gdpconfig.is_dir():
#    # find vasp config
#    vasprc = gdpconfig / "vasprc.json"
#    with open(vasprc, "r") as fopen:
#        input_dict = json.load(fopen)
#else:
#    input_dict = {}

#: Model deviations by the committee model.
VALID_DEVI_FRAME_KEYS: List[str] = [
    "devi_te",
    "max_devi_v", "min_devi_v", "avg_devi_v",
    "max_devi_f", "min_devi_f", "avg_devi_f",
    "max_devi_ae", "min_devi_ae", "avg_devi_ae",
]

#: Model deviations by the committee model.
VALID_DEVI_ATOMIC_KEYS: List[str] = [
    "devi_f",
]

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
plt.style.use("presentation")

from scipy.interpolate import griddata, interp1d
from scipy.optimize import minimize

# Load and reshape the data
#data = np.loadtxt("data.txt")
data = np.loadtxt("fes.dat")
print(f"data: {data.shape}")
data[:, 2] /= 96.485

# Guesses for where the basins go. Since the input data is malformed, 
# you need to also provide an intermediate point t.
#a = (-0.05, 0.25)
#b = (0.3, 0.25)
#t = (0.2, 0.1) #Transition basin
#a = (0.00, 0.05)
#b = (0.33, 0.05)
#t = (0.17, 0.05) #Transition basin
#a = (0.06, 0.00)
#b = (0.06, 0.50)
#t = (0.06, 0.25) #Transition basin
a = (0.28, 0.05)
b = (0.33, 0.05)
t = (0.30, 0.05) #Transition basin
npts = 10 #Number of intermediate points on the string.

nx = np.size(np.unique(data[:, 0]))
ny = np.size(np.unique(data[:, 1]))

# Some people give input where the first column varies fastest. 
# That is Fortran ordering, and you are liable to get things confused 
# if you don't take this into account.
order = "C"
if data[0, 0] != data[1, 0]:
	order = "F"
x = data[:, 0].reshape(nx, ny, order=order)
y = data[:, 1].reshape(nx, ny, order=order)
z = data[:, 2].reshape(nx, ny, order=order)

# Basic input sanity check.
xdiff = np.diff(x, axis=0)
ydiff = np.diff(y, axis=1)
if (not np.all(np.abs(xdiff - xdiff[0]) < 1e-8)) or (not np.all(np.abs(ydiff - ydiff[0]) < 1e-8)):
	print("WARNING! The input data is not coming from an equally spaced grid. imshow will be subtly wrong as a result, as each pixel is assumed to represent the same area.")

gridpoints = np.vstack([x.flatten(), y.flatten()]).T
print(f"grids: {gridpoints.shape}")

# Make a figure and a axes
fig, ax = plt.subplots(1, 1, figsize=(12, 8))

# Use the guessed basins to make the original string, 
# a linear interpolation between points a, t, and b.
pts = np.vstack(
	[
		np.hstack([np.linspace(a[0], t[0], npts), np.linspace(t[0], b[0], npts)]),
		np.hstack([np.linspace(a[1], t[1], npts), np.linspace(t[1], b[1], npts)])
	]
).T
gradx, grady = np.gradient(z, np.unique(data[:,0]), np.unique(data[:,1]))
# Evolve points so that they respond to the gradients. 
# This is the "zero temperature string method"
stepmax = 100
for i in range(stepmax):
	#Find gradient interpolation
	Dx = griddata(gridpoints,gradx.flatten(),(pts[:,0],pts[:,1]), method='linear')
	Dy = griddata(gridpoints,grady.flatten(),(pts[:,0],pts[:,1]), method='linear')
	h = np.amax(np.sqrt(np.square(Dx)+np.square(Dy)))
	# Evolve
	pts -= 0.01 * np.vstack([Dx,Dy]).T / h
	# Reparameterize
	arclength = np.hstack([0,np.cumsum(np.linalg.norm(pts[1:] - pts[:-1],axis=1))])
	arclength /= arclength[-1]
	pts = np.vstack([interp1d(arclength,pts[:,0])(np.linspace(0,1,2*npts)), interp1d(arclength,pts[:,1])(np.linspace(0,1,2*npts))]).T
	if i % 10 == 0:
		print(i, np.sum(griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear")))
		# This draws the intermediate states to visualize how the string evolves.
		#ax.plot(pts[:, 0], pts[:, 1], color=plt.cm.spring(i/float(stepmax)))
		ax.plot(pts[:, 0], pts[:, 1], ls="--")

np.savetxt("./path.txt", pts, fmt="%.2f")
ax.plot(pts[:, 0], pts[:,1], color="k", linestyle="--") # converged pathway?

#heatmap = ax.imshow(
#	z.T, cmap=plt.cm.rainbow, vmin = np.nanmin(z), 
#	vmax=1.2 * np.nanmax(griddata(gridpoints,z.flatten(),(pts[:,0],pts[:,1]), method='linear')), 
#	origin='lower', aspect='auto', extent = (x[0][0], x[-1][-1],y[0][0], y[-1][-1]), 
#	interpolation="bicubic"
#)
#heatmap.cmap.set_over('white')
#ax.autoscale(False)
#bar = fig.colorbar(heatmap)
#bar.set_label("Free Energy [eV]", rotation=90, fontsize=12)
ax.set_xlabel("CV 1")
ax.set_ylabel("CV 2")
ax.set_xlim(0., 0.33)
ax.set_ylim(0., 0.50)
#cntr = ax.contourf(
#    x, y, z, levels=10,
#    #x, np.where(y < 0., y+1.0, y), z, levels=10,
#    #cmap="RdBu"
#)
#fig.colorbar(cntr, ax=ax, label="Free Energy [eV]")
#cn = ax.contour(
#    x, y, z, #cmap="RdBu"
#    #x, np.where(y < 0., y, y+1.0), z, levels=10,
#)
#plt.clabel(cn, inline=True, fontsize=10)

#ax.xaxis.set_ticks([1.25,1.50,1.75,2.0,2.25,2.50,2.75,3.0,3.25])
#ax.yaxis.set_ticks([1.25,1.50,1.75,2.0,2.25,2.50,2.75])
fig.savefig("demo.png", transparent=True)
#FYI this can also save .pdf or .svg or .eps, if you want to go in a vectorly direction.

# --- 
fig, ax = plt.subplots(1, 1, figsize=(12, 8))
ax.plot(
	np.linspace(0, 1, 2*npts), 
	griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear"),
    marker="o"
)
ax.set_ylabel("Free Energy [eV]")
ax.set_xlabel("Reaction Coordinate")
fig.savefig("pmf.png")

if __name__ == "__main__":
	...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import copy
import dataclasses
import pathlib
import subprocess
from typing import Optional, Union, List, Tuple, Callable

import json

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator

from .. import AbstractPotentialManager, AbstractTrainer
from .. import DummyCalculator, CommitteeCalculator

from .convert import convert_groups


@dataclasses.dataclass
class DeepmdSystem:

    #: System name.
    name: str

    #: Frames.
    frames: List[Atoms]

    #: Number of train_and_split.
    train_and_split: Tuple[int, int]

    #: Batchsize
    batchsize: int = 1

    def __post_init__(
        self,
    ):
        """"""
        self.nframes = len(self.frames)
        assert self.nframes == sum(
            self.train_and_split
        ), f"{self.name}: {self.nframes} != sum({self.train_and_split})"

        composition_list = [a.get_chemical_formula() for a in self.frames]
        assert len(set(composition_list)) == 1, f"{self.name}: {composition_list[0]}?"
        self.composition = composition_list[0]

        return

    def write(self, directory: pathlib.Path) -> None:
        """"""
        directory.mkdir(parents=True, exist_ok=True)
        write(directory / f"{self.composition}.xyz", self.frames)

        return


class DeepmdDataloader:

    #: Datasset name.
    name: str = "deepmd"

    #:
    _systems: List[DeepmdSystem] = []

    def __init__(
        self,
        batchsizes,
        cum_batchsizes,
        train_sys_dirs,
        valid_sys_dirs,
        *args,
        **kwargs,
    ) -> None:
        """"""
        self.batchsizes = batchsizes
        self.cum_batchsizes = cum_batchsizes
        self.train_sys_dirs = [str(x) for x in train_sys_dirs]
        self.valid_sys_dirs = [str(x) for x in valid_sys_dirs]

        return

    @staticmethod
    def from_directory(
        train_directory: Union[str, pathlib.Path],
        valid_directory: Optional[Union[str, pathlib.Path]]=None,
    ) -> "DeepmdDataloader":
        """"""
        # - read trainset...
        train_wdir = pathlib.Path(train_directory)
        trainset_dirs = sorted(train_wdir.iterdir())

        if valid_directory is not None:
            valid_wdir = pathlib.Path(valid_directory)
            validset_dirs = sorted(valid_wdir.iterdir())
        else:
            validset_dirs = []

        batchsizes = [1]*len(trainset_dirs)
        cum_batchsizes = sum(batchsizes)

        return DeepmdDataloader(batchsizes, cum_batchsizes, trainset_dirs, validset_dirs)

    @property
    def systems(
        self,
    ):
        """"""

        return self._systems

    def load(
        self,
    ) -> None:
        """"""
        systems = []
        for p in self.train_sys_dirs:
            p = pathlib.Path(p)
            print(p)
            curr_train_frames = DeepmdDataloader.convert_system_to_frames(p)
            num_curr_train_frames = len(curr_train_frames)
            curr_frames = curr_train_frames
            for p2 in self.valid_sys_dirs:
                p2 = pathlib.Path(p2)
                if p2.name == p.name:
                    curr_valid_frames = DeepmdDataloader.convert_system_to_frames(p2)
                    num_curr_valid_frames = len(curr_valid_frames)
                    curr_frames.extend(curr_valid_frames)
                    break
            else:
                num_curr_valid_frames = 0
            curr_system = DeepmdSystem(
                p.name, curr_frames, (num_curr_train_frames, num_curr_valid_frames)
            )
            systems.append(curr_system)
        self._systems = systems

        return

    def dump(self, directory: Union[str, pathlib.Path]) -> None:
        """"""
        directory = pathlib.Path(directory)
        for curr_system in self.systems:
            curr_system.write(
                directory / (curr_system.name + "-" + curr_system.composition),
            )
            print(
                f"{curr_system.name =} {curr_system.composition} {curr_system.train_and_split} {curr_system.nframes}"
            )

        return

    @staticmethod
    def set2frames(
        set_dir: pathlib.Path, chemical_symbols: List[str], pbc: List[int]
    ) -> List[Atoms]:
        """Convert set into frames."""
        boxes = np.load(set_dir / "box.npy")
        nframes = boxes.shape[0]
        coords = np.load(set_dir / "coord.npy")
        energies = np.load(set_dir / "energy.npy")
        forces = np.load(set_dir / "force.npy")

        if nframes == 1:
            energies = energies.reshape(-1)

        frames = []
        for i in range(nframes):
            cell = boxes[i, :].reshape(3, 3)
            positions = coords[i, :].reshape(-1, 3)
            atoms = Atoms(
                symbols=chemical_symbols, positions=positions, cell=cell, pbc=pbc
            )
            results = {"energy": energies[i], "forces": forces[i, :].reshape(-1, 3)}
            spc = SinglePointCalculator(atoms, **results)
            atoms.calc = spc
            frames.append(atoms)

        return frames

    @staticmethod
    def convert_system_to_frames(curr_system: pathlib.Path):
        """"""
        # find all set dirs
        set_dirs = sorted(curr_system.glob("set*"))

        type_list = np.loadtxt(curr_system / "type_map.raw", dtype=str)
        if len(type_list.shape) == 0:
            type_list = type_list.reshape(-1)
        atype = np.loadtxt(curr_system / "type.raw", dtype=int)
        if len(atype.shape) == 0:
            atype = atype.reshape(-1)
        chemical_symbols = [type_list[a] for a in atype]
        pbc = [1, 1, 1] if not (curr_system / "nopbc").exists() else [0, 0, 0]

        # train data
        frames = []
        for set_dir in set_dirs[:]:
            frames.extend(DeepmdDataloader.set2frames(set_dir, chemical_symbols, pbc))

        return frames

    def as_dict(
        self,
    ) -> dict:
        """"""
        params = {}
        params["name"] = self.name
        params["batchsizes"] = self.batchsizes
        params["cum_batchsizes"] = self.cum_batchsizes
        params["train_sys_dirs"] = self.train_sys_dirs
        params["valid_sys_dirs"] = self.valid_sys_dirs

        return params


class DeepmdTrainer(AbstractTrainer):

    name = "deepmd"
    command = "dp"
    freeze_command = "dp"
    prefix = "config"

    #: Flag indicates that the training is finished properly.
    CONVERGENCE_FLAG: str = "finished training"

    def __init__(
        self,
        config: dict,
        type_list: List[str] = None,
        train_epochs: int = 200,
        print_epochs: int = 5,
        directory=".",
        command="dp",
        freeze_command="dp",
        random_seed=1112,
        *args,
        **kwargs,
    ) -> None:
        """"""
        super().__init__(
            config=config,
            type_list=type_list,
            train_epochs=train_epochs,
            print_epochs=print_epochs,
            directory=directory,
            command=command,
            freeze_command=freeze_command,
            random_seed=random_seed,
            *args,
            **kwargs,
        )

        # - TODO: sync type_list
        if type_list is None:
            self._type_list = config["model"]["type_map"]
        else:
            self._type_list = type_list

        return

    def _resolve_train_command(self, init_model=None):
        """"""
        train_command = self.command

        # - add options
        command = "{} train {}.json ".format(train_command, self.name)
        if init_model is not None:
            init_model_path = pathlib.Path(init_model).resolve()
            if init_model_path.name.endswith(".pb"):
                command += "--init-frz-model {}".format(str(init_model_path))
            elif init_model_path.name.endswith("model.ckpt"):
                command += "--init-model {}".format(str(init_model_path))
            else:
                raise RuntimeError(f"Unknown init_model {str(init_model_path)}.")
        command += " 2>&1 > {}.out".format(self.name)

        return command

    def _resolve_freeze_command(self, *args, **kwargs):
        """"""
        freeze_command = self.command

        # - add options
        command = "{} freeze -o {} 2>&1 >> {}.out".format(
            freeze_command, self.frozen_name, self.name
        )

        return command

    def _resolve_compress_command(self, *args, **kwargs):
        """"""
        compress_command = self.command

        # - add options
        command = "{} compress -i {} -o {} 2>&1 >> {}.out".format(
            compress_command, self.frozen_name, f"{self.name}-c.pb", self.name
        )

        return command

    @property
    def frozen_name(self):
        """"""
        return f"{self.name}.pb"

    def _train_from_the_restart(self, dataset, init_model):
        """Train from the restart"""
        if not self.directory.exists():
            command = self._train_from_the_scratch(dataset, init_model)
        else:
            ckpt_info = self.directory / "checkpoint"
            if ckpt_info.exists() and ckpt_info.stat().st_size != 0:
                # TODO: check if the ckpt model exists?
                command = f"{self.command} train {self.name}.json "
                command += f"--restart model.ckpt"
                self._print(f"TRAINING COMMAND: {command}")
            else:  # assume not at any ckpt so start from the scratch
                command = self._train_from_the_scratch(dataset, init_model)

        return command

    def _prepare_dataset(self, dataset, *args, **kwargs):
        """"""
        if not self.directory.exists():
            self.directory.mkdir(parents=True, exist_ok=True)
        if not isinstance(dataset, DeepmdDataloader):
            set_names, train_frames, test_frames, adjusted_batchsizes = (
                dataset.split_train_and_test()
            )
            train_dir = self.directory

            # - update config
            self._print("--- write dp train data---")
            batchsizes = adjusted_batchsizes
            cum_batchsizes, train_sys_dirs = convert_groups(
                set_names,
                train_frames,
                batchsizes,
                self.config["model"]["type_map"],
                "train",
                train_dir,
                self._print,
            )
            _, valid_sys_dirs = convert_groups(
                set_names,
                test_frames,
                batchsizes,
                self.config["model"]["type_map"],
                "valid",
                train_dir,
                self._print,
            )
            self._print(f"accumulated number of batches: {cum_batchsizes}")

            dataset = DeepmdDataloader(
                batchsizes, cum_batchsizes, train_sys_dirs, valid_sys_dirs
            )
        else:
            ...

        return dataset

    def write_input(self, dataset):
        """Write inputs for training."""
        # - prepare dataset (convert dataset to DeepmdDataloader)
        dataset = self._prepare_dataset(dataset)

        # - check train config
        # NOTE: parameters
        #       numb_steps, seed
        #       descriptor-seed, fitting_net-seed
        #       training - training_data, validation_data
        train_config = copy.deepcopy(self.config)

        train_config["model"]["descriptor"]["seed"] = self.rng.integers(
            0, 10000, dtype=int
        )
        train_config["model"]["fitting_net"]["seed"] = self.rng.integers(
            0, 10000, dtype=int
        )

        train_config["training"]["training_data"]["systems"] = [
            x for x in dataset.train_sys_dirs
        ]
        train_config["training"]["training_data"]["batch_size"] = dataset.batchsizes

        train_config["training"]["validation_data"]["systems"] = [
            x for x in dataset.valid_sys_dirs
        ]
        train_config["training"]["validation_data"]["batch_size"] = dataset.batchsizes

        train_config["training"]["seed"] = self.rng.integers(0, 10000, dtype=int)

        # --- calc numb_steps
        min_freq_unit = 100.0
        save_freq = int(
            np.ceil(dataset.cum_batchsizes * self.print_epochs / min_freq_unit)
            * min_freq_unit
        )
        train_config["training"]["save_freq"] = save_freq

        numb_steps = dataset.cum_batchsizes * self.train_epochs
        n_checkpoints = int(
            np.ceil(dataset.cum_batchsizes * self.train_epochs / save_freq)
        )
        numb_steps = n_checkpoints * save_freq
        train_config["training"]["numb_steps"] = numb_steps

        # - write
        with open(self.directory / f"{self.name}.json", "w") as fopen:
            json.dump(train_config, fopen, indent=2)

        return

    def freeze(self):
        """"""
        # - freeze model
        frozen_model = super().freeze()

        # - compress model
        compressed_model = (self.directory / f"{self.name}-c.pb").absolute()
        if frozen_model.exists() and not compressed_model.exists():
            command = self._resolve_compress_command()
            try:
                proc = subprocess.Popen(command, shell=True, cwd=self.directory)
            except OSError as err:
                msg = "Failed to execute `{}`".format(command)
                # raise RuntimeError(msg) from err
                # self._print(msg)
                self._print("Failed to compress model.")
            except RuntimeError as err:
                self._print("Failed to compress model.")

            errorcode = proc.wait()
            if errorcode:
                path = os.path.abspath(self.directory)
                msg = (
                    'Trainer "{}" failed with command "{}" failed in '
                    "{} with error code {}".format(self.name, command, path, errorcode)
                )
                # NOTE: sometimes dp cannot compress the model
                #       this happens when the descriptor trainable is set False?
                # raise RuntimeError(msg)
                # self._print(msg)
                compressed_model.symlink_to(
                    frozen_model.relative_to(compressed_model.parent)
                )
        else:
            ...

        return compressed_model

    def read_convergence(self) -> bool:
        """Read training convergence.

        Check deepmd training progress by comparing the `numb_steps` in the input
        configuration and the current step in `lcurve.out`.

        """
        self._print(f"check {self.name} training convergence...")
        converged = False

        dpconfig_path = self.directory / f"{self.name}.json"
        if dpconfig_path.exists():
            # - get numb_steps
            with open(dpconfig_path, "r") as fopen:
                input_json = json.load(fopen)
            numb_steps = input_json["training"]["numb_steps"]

            # - get current step
            lcurve_out = self.directory / f"lcurve.out"
            if lcurve_out.exists():
                with open(lcurve_out, "r") as fopen:
                    lines = fopen.readlines()
                try:
                    curr_steps = int(lines[-1].strip().split()[0])
                    if curr_steps >= numb_steps:
                        converged = True
                    self._debug(f"{curr_steps} >=? {numb_steps}")
                except:
                    self._print(f"The endline of `lcure.out` is strange.")
            else:
                ...
        else:
            ...

        return converged


class DeepmdManager(AbstractPotentialManager):

    name = "deepmd"

    implemented_backends = ["ase", "lammps"]

    valid_combinations = (
        # calculator, dynamics
        ("ase", "ase"),
        ("lammps", "ase"),
        ("lammps", "lammps"),
    )

    #: Used for estimating uncertainty.
    _estimator = None

    def __init__(self, *args, **kwargs):
        """"""

        return

    def _create_calculator(self, calc_params: dict) -> Calculator:
        """Create an ase calculator.

        Todo:
            In fact, uncertainty estimation has various backends as well.

        """
        calc_params = copy.deepcopy(calc_params)

        # - some shared params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        type_list = calc_params.pop("type_list", [])
        type_map = {}
        for i, a in enumerate(type_list):
            type_map[a] = i

        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))
        self.calc_params.update(model=models)

        # TODO: make this a dataclass??
        #       currently, default disable uncertainty estimation
        estimate_uncertainty = calc_params.get("estimate_uncertainty", False)

        # - create specific calculator
        calc = DummyCalculator()
        if self.calc_backend == "ase":
            # return ase calculator
            try:
                from .calculator import DP
            except:
                raise ModuleNotFoundError(
                    "Please install deepmd-kit to use the ase interface."
                )
            # if models and type_map:
            #    calc = DP(model=models[0], type_dict=type_map)
            calcs = []
            for m in models:
                curr_calc = DP(model=m, type_dict=type_map)
                calcs.append(curr_calc)
            if len(calcs) == 1:
                calc = calcs[0]
            elif len(calcs) > 1:
                if estimate_uncertainty:
                    calc = CommitteeCalculator(calcs=calcs)
                else:
                    calc = calcs[0]
            else:
                ...
        elif self.calc_backend == "lammps":
            from gdpx.computation.lammps import Lammps

            if models:
                if len(models) == 1:
                    pair_style = "deepmd {}".format(" ".join(models))
                else:
                    if estimate_uncertainty:
                        pair_style = "deepmd {}".format(" ".join(models))
                    else:
                        pair_style = "deepmd {}".format(models[0])
                pair_coeff = calc_params.pop("pair_coeff", "* *")

                pair_style_name = pair_style.split()[0]
                assert (
                    pair_style_name == "deepmd"
                ), "Incorrect pair_style for lammps deepmd..."

                calc = Lammps(
                    command=command,
                    directory=directory,
                    pair_style=pair_style,
                    pair_coeff=pair_coeff,
                    **calc_params,
                )
                # - update several params
                calc.units = "metal"
                calc.atom_style = "atomic"

        return calc

    def register_calculator(self, calc_params, *args, **kwargs) -> None:
        """generate calculator with various backends"""
        super().register_calculator(calc_params)

        self.calc = self._create_calculator(self.calc_params)

        return

    def switch_backend(self, backend: str = None) -> None:
        """Switch the potential's calculation backend."""
        if backend is None:
            return

        if not hasattr(self, "calc"):
            raise RuntimeError(
                f"{self.name} cannot switch backend as it does not have a calculator attached."
            )
        if backend not in self.implemented_backends:
            raise RuntimeError(
                f"{self.name} cannot switch backend from {self.calc_backend} to {backend}."
            )

        prev_backend = self.calc_backend
        if prev_backend == "ase" and backend == "lammps":
            calc_params = copy.deepcopy(self.calc_params)
            calc_params["backend"] = "lammps"
            command = calc_params.get("command", None)
            if command is None:
                raise RuntimeError(
                    f"{self.name} cannot switch backend from ase to lammps as no command is provided."
                )
            else:
                self.calc_backend = None
            self.register_calculator(calc_params)
        elif prev_backend == "lammps" and backend == "ase":
            calc_params = copy.deepcopy(self.calc_params)
            calc_params["backend"] = "ase"
            self.calc_backend = None
            self.register_calculator(calc_params)
        else:  # Nothing to do for other combinations
            ...

        return

    def switch_uncertainty_estimation(self, status: bool = True):
        """Switch on/off the uncertainty estimation."""
        # NOTE: Sometimes the manager loads several models and supports uncertainty
        #       by committee but the user disables it. We need change the calc to
        #       the correct one as the loaded one is just a single calculator.
        if not hasattr(self, "calc"):
            raise RuntimeError(
                "Fail to switch uncertainty status as it does not have a calc."
            )
        # print(f"{self.calc}")

        # NOTE: make sure manager.as_dict() can have correct param
        self.calc_params["estimate_uncertainty"] = status

        # - convert calculator
        if self.calc_backend == "ase":
            if status:
                if isinstance(self.calc, CommitteeCalculator):
                    ...  # nothing to do
                else:  # reload models
                    self.calc = self._create_calculator(self.calc_params)
            else:
                if isinstance(self.calc, CommitteeCalculator):
                    # TODO: save previous calc?
                    self.calc = self.calc.calcs[0]
                else:
                    ...
        elif self.calc_backend == "lammps":
            models = self.calc.pair_style.split()[1:]  # model paths
            nmodels = len(models)
            if status:
                if nmodels > 1:
                    ...
                else:
                    self.calc = self._create_calculator(self.calc_params)
            else:
                # TODO: use self.calc_params? It should be protected?
                # pair_style deepmd m0 m1 m2 m3
                if nmodels > 1:
                    self.calc.pair_style = f"deepmd {models[0]}"
                else:
                    ...
        else:
            # TODO:
            # Other backends cannot have uncertainty estimation,
            # give a warning?
            ...
        # print(f"{self.calc}")

        return

    def remove_loaded_models(self, *args, **kwargs):
        """Loaded TF models should be removed before any copy.deepcopy operations."""
        self.calc.reset()
        if self.calc_backend == "ase":
            if isinstance(self.calc, CommitteeCalculator):
                for c in self.calc.calcs:
                    c.dp = None
            else:
                self.calc.dp = None
        else:
            ...

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Generate various defects using
pymatgen.analysis.defects.generators module
and
https://github.com/kaist-amsg/LS-CGCNN-ens/blob/master/FindSite.py
"""

from scipy.spatial import Delaunay
import numpy as np
from collections import defaultdict
import copy

from ase.io import read, write
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.defects.generators import InterstitialGenerator, VoronoiInterstitialGenerator

def alpha_shape_3D(pos, alpha,roundtol = 9):
    """
    Compute the alpha shape (concave hull) of a set of 3D points.
    Parameters:
        pos - np.array of shape (n,3) points.
        alpha - alpha value.
        roundtol - number of decimal for rounding 
    return
        outer surface vertex indices, edge indices, and triangle indices
    """

    tetra_vertices = Delaunay(pos).vertices
    # Find radius of the circumsphere.
    # By definition, radius of the sphere fitting inside the tetrahedral needs 
    # to be smaller than alpha value
    # http://mathworld.wolfram.com/Circumsphere.html
    tetrapos = np.take(pos,tetra_vertices,axis=0)
    normsq = np.sum(tetrapos**2,axis=2)[:,:,None]
    ones = np.ones((tetrapos.shape[0],tetrapos.shape[1],1))
    a = np.linalg.det(np.concatenate((tetrapos,ones),axis=2))
    Dx = np.linalg.det(np.concatenate((normsq,tetrapos[:,:,[1,2]],ones),axis=2))
    Dy = -np.linalg.det(np.concatenate((normsq,tetrapos[:,:,[0,2]],ones),axis=2))
    Dz = np.linalg.det(np.concatenate((normsq,tetrapos[:,:,[0,1]],ones),axis=2))
    c = np.linalg.det(np.concatenate((normsq,tetrapos),axis=2))
    # Remove bad tetrahedrals. These the ones where volume is zero.
    bad = a==0
    num = Dx**2+Dy**2+Dz**2-4*a*c
    bad[num<0] = True
    bad = np.where(bad)[0]
    tetra_vertices = np.delete(tetra_vertices,bad,axis=0)
    num = np.delete(num,bad,axis=0)
    a = np.delete(a,bad,axis=0)
    # get radius
    r = np.sqrt(num)/(2*np.abs(a))

    # Find tetrahedrals
    tetras = tetra_vertices[r<alpha,:]
    # triangles
    TriComb = np.array([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)])
    Triangles = tetras[:,TriComb].reshape(-1,3)
    Triangles = np.sort(Triangles,axis=1)
    # Remove triangles that occurs twice, because they are within shapes
    TrianglesDict = defaultdict(int)
    for tri in Triangles:TrianglesDict[tuple(tri)] += 1
    Triangles=np.array([tri for tri in TrianglesDict if TrianglesDict[tri] ==1])
    #edges
    if Triangles.size==0:
        return [], [], []
    EdgeComb=np.array([(0, 1), (0, 2), (1, 2)])
    Edges=Triangles[:,EdgeComb].reshape(-1,2)
    Edges=np.sort(Edges,axis=1)
    Edges=np.unique(Edges,axis=0)

    Vertices = np.unique(Edges)
    return Vertices,Edges,Triangles

atoms = read('/users/40247882/scratch2/validations/surfaces/surf-111.xyz') # (2x2) Pt(111) surface
#print(atoms)

structure = AseAtomsAdaptor.get_structure(atoms)
#print(structure)

vig = InterstitialGenerator(structure, 'O')
print(vig.unique_defect_seq)
exit()
structures = [v.generate_defect_structure() for v in vig]
frames = [AseAtomsAdaptor.get_atoms(s) for s in structures]

write('miaow.xyz', frames)


if __name__ == '__main__':
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy

from .. import config
from ..core.register import registers

from ..builder.constraints import parse_constraint_info
from ..potential.calculators.mixer import EnhancedCalculator

from ..utils.strucopy import read_sort, resort_atoms_with_spc
from ..utils.cmdrun import run_ase_calculator


""" This submodule is for exploring, sampling, 
    and performing (chemical) reactions with
    various advanced algorithms.
"""

# - string methods...
from .string import (
    AseStringReactor, Cp2kStringReactor, VaspStringReactor,
    ZeroStringReactor
)
registers.reactor.register("ase")(AseStringReactor)
registers.reactor.register("cp2k")(Cp2kStringReactor)
registers.reactor.register("vasp")(VaspStringReactor)
registers.reactor.register("grid")(ZeroStringReactor)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import json
import argparse
from itertools import groupby

import numpy as np

"""
Following the below protocol, 
we analyse and reduce the size of dataset.

1. Use SOAP to represent each structure.
2. Use HDBSCAN to cluster structures.
3. Use histogram and CUR to select most representative structures.
"""

from sklearn.datasets import make_blobs
# import pandas as pd

import hdbscan

from ase.io import read, write

from dscribe.descriptors import SOAP

from gdpx.utils.data.cur import cur_selection

def frames2features(frames, feature_computer, njobs=1, saved_file=None):
    """ use dscribe to calculate structure representation
    """
    print("feature dimension: ", feature_computer.get_number_of_features())

    # computation
    features = feature_computer.create(frames, n_jobs=njobs)

    # save calculated features 
    if saved_file is not None:
        np.save(saved_file, features)

    assert len(frames) == features.shape[0], "number of frames != number of features"

    return features


def simple_test():
    blobs, labels = make_blobs(n_samples=2000, n_features=10)
    clusterer = hdbscan.HDBSCAN()

    """
    HDBSCAN(
        algorithm='best', alpha=1.0, approx_min_span_tree=True,
        gen_min_span_tree=False, leaf_size=40, memory=Memory(cachedir=None),
        metric='euclidean', # different distance metrics
        min_cluster_size=5, # minimum number of samples in each cluster
        min_samples=None, # number of samples to form dense area, default is min_cluster_size
        p=None
    )
    """

    clusterer.fit(blobs)
    print(clusterer.labels_)

    # clusterer.probabilities_ # each sample has probs of each cluster

def cluster_features(features, frames):
    """"""
    print("===== Start Clustering =====")
    clusterer = hdbscan.HDBSCAN(
        #min_cluster_size = 32
        min_cluster_size = 8 # avoid too many small clusters
    )

    clusterer.fit(features)
    #print(clusterer.labels_)
    nclusters = len(set(clusterer.labels_)) - 1 # -1 means noises
    print("number of clusters: ", nclusters)
    #print(clusterer.probabilities_) # each sample has probs of each cluster

    #frame_clusters = []
    #for i in range(nclusters+1):
    #    frame_clusters.append([])
    #print(frame_clusters)

    index_clusters = []
    for i in range(nclusters+1):
        index_clusters.append([])
    #print(index_clusters)

    for i, idx in zip(clusterer.labels_, range(len(frames))):
        # print(i)
        index_clusters[i].append(idx)

    return index_clusters

def cluster_and_cur(frames, feature_cached=True):
    # read frames
    #frames = read("/users/40247882/scratch2/analyse-data/O24Pt24-test.xyz", ":")

    # calculate features
    with open("/users/40247882/scratch2/analyse-data/soap_param.json", 'r') as fopen:
        input_dict = json.load(fopen)

    soap = SOAP(**input_dict['soap'])
    if feature_cached:
        features = np.load("fea.npy")
        print("features are loaded from file")
    else:
        start_time = time.time()
        features = frames2features(frames, soap, njobs=4, saved_file="fea.npy")
        end_time = time.time()
        print("feature calculation time: ", end_time - start_time)

    # clustering
    start_time = time.time()
    index_clusters = cluster_features(features, frames)
    end_time = time.time()
    print("clustering time: ", end_time - start_time)
    #print(index_clusters)

    #for i in range(-1, nclusters):
    #    print("#frames %d in cluster %d" %(len(frame_clusters[i]), i))
    #    write("xxx-%d.xyz" %i, frame_clusters[i])

    for icluster, indices in enumerate(index_clusters):
        print("===== Cluster %d =====" %icluster)
        cur_frames = []
        for x in indices:
            atoms = frames[x]
            atoms.info["index"] = x
            cur_frames.append(atoms)
        write("group-"+str(icluster)+".xyz", cur_frames)
        print("Current #frames: ", len(indices))
    
    return


def reduce_group(frames, features, cluster_indices):
    """ reduce dataset by CUR decomposition
    """
    index_clusters = []
    for i in cluster_indices:
        cur_frames = read("./groups/group-"+str(i)+".xyz", ":")
        indices = [atoms.info["index"] for atoms in cur_frames]
        index_clusters.append(indices)
    # CUR decomposition for dataset reduction
    min_size = 500
    ratio = 0.5 # dataset reduction ratio
    for icluster, indices in zip(cluster_indices, index_clusters):
        print("cluster group ", icluster)
        cname = "group-"+str(icluster)+"_cured.xyz"
        if len(indices) > min_size:
            cur_features = features[indices,:]
            cur_nframes = cur_features.shape[0]
            print("Current #frames", cur_nframes)
            nselected = int(np.ceil(cur_nframes*ratio))
            print("Current #selected", nselected)
            cur_scores, cur_selected = cur_selection(cur_features, nselected, zeta=-1, strategy="descent")
            selected_indices = [indices[x] for x in cur_selected]
            # TODO: add index info to selected frames
            cur_frames = [frames[x] for x in selected_indices]
            write(cname, cur_frames)
        else:
            print("Clustersize is too small only with %d" %len(indices))
            continue
        
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--frames", nargs="*",
        required=True,
        help="input frames"
    )
    parser.add_argument(
        "--fea", action="store_true",
        help="if features are calculated"
    )

    parser.add_argument(
        "--cur", action="store_true",
        help="cur decomposition"
    )
    parser.add_argument(
        "-ci", "--cluster_indices", nargs="*",
        help="cur decomposition"
    )

    args = parser.parse_args()

    frames = []
    for f in args.frames:
        frames.extend(
            read(f, ":")
        )
    print("#frames: ", len(frames))

    if not args.cur:
        # clustering
        cluster_and_cur(frames, args.fea)
    else:
        features = np.load('fea.npy')
        reduce_group(frames, features, args.cluster_indices)


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import pathlib
from typing import NoReturn

from . import AbstractPotentialManager


class EmtManager(AbstractPotentialManager):

    name = "emt"

    implemented_backends = ["emt", "ase"]
    valid_combinations = (
        ("ase", "ase"), 
    )

    """See ASE documentation for calculator parameters.
    """

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)
        if self.calc_backend == "emt":
            self.calc_backend = "ase"

        # NOTE: emt backend is just an alias of ase backend, they are the same.
        if self.calc_backend == "ase":
            from ase.calculators.emt import EMT
            calc_cls = EMT
        else:
            raise NotImplementedError(f"Unsupported backend {self.calc_backend}.")
        
        calc = calc_cls(**calc_params)

        self.calc = calc

        return


if __name__ == "__main__":
    ...



#!/usr3/bin/env python3
# -*- coding: utf-8 -*

import copy
import os
import pathlib
import shutil

from typing import List

import numpy as np

from ase import units
from ase.units import fs, mol, kJ, nm
from ase.calculators.calculator import Calculator, all_changes

try:
    import plumed
except Exception as e:
    print(e)


"""A plumed wrapper for ase.

There is an official plumed calculator in ase master branch but not in v3.22.1 that
can be installed from conda.

Units setup
warning: inputs and outputs of plumed will still be in
plumed units.

The change of Plumed units to ASE units is:
kjoule/mol to eV
nm to Angstrom
ps to ASE time units
ASE and plumed - charge unit is in e units
ASE and plumed - mass unit is in a.m.u units

Notes:

    conda install -c conda-forge plumed
    conda install -c conda-forge py-plumed

"""


def update_input_value(line: str, key: str, value, func: callable):
    """Update the given key with the new value."""
    shift = len(key) + 1  # key name and =
    if line.find(key) != -1:
        ini = line.find(key)
        end = line.find(" ", ini)
        if end == -1:
            prev = line[ini + shift :]
            line = line[: ini + shift] + func(prev, value)
        else:
            prev = line[ini + shift : end]
            line = line[: ini + shift] + func(prev, value) + line[end:]

    return line


def update_stride_and_file(input_lines: List[str], wdir: str, stride: int) -> List[str]:
    """"""
    input_lines, parsed_lines = copy.deepcopy(input_lines), []
    for line in input_lines:
        parsed_line = update_input_value(
            line, "FILE", wdir, func=lambda x, y: os.path.join(y, x)
        )
        parsed_line = update_input_value(
            parsed_line, "STRIDE", stride, func=lambda x, y: str(y)
        )
        parsed_lines.append(parsed_line)

    return parsed_lines


class Plumed(Calculator):

    implemented_properties = ["energy", "forces"]

    def __init__(
        self,
        input: List[str],
        atoms=None,
        kT=1.0,
        restart=False,
        use_charge=False,
        update_charge=False,
    ):
        """
        Plumed calculator is used for simulations of enhanced sampling methods
        with the open-source code PLUMED (plumed.org).

        [1] The PLUMED consortium, Nat. Methods 16, 670 (2019)
        [2] Tribello, Bonomi, Branduardi, Camilloni, and Bussi,
        Comput. Phys. Commun. 185, 604 (2014)

        Parameters
        ----------
        input: List of strings
            It contains the setup of plumed actions

        atoms: Atoms
            Atoms object to be attached

        kT: float. Default 1.
            Value of the thermal energy in eV units. It is important for
            some methods of plumed like Well-Tempered Metadynamics.

        restart: boolean. Default False
            True if the simulation is restarted.

        use_charge: boolean. Default False
            True if you use some collective variable which needs charges. If
            use_charges is True and update_charge is False, you have to define
            initial charges and then this charge will be used during all
            simulation.

        update_charge: boolean. Default False
            True if you want the charges to be updated each time step. This
            will fail in case that calc does not have 'charges' in its
            properties.

        .. note:: For this case, the calculator is defined strictly with the
            object atoms inside. This is necessary for initializing the
            Plumed object. For conserving ASE convention, it can be initialized
            as atoms.calc = (..., atoms=atoms, ...)


        .. note:: In order to guarantee a proper restart, the user has to fix
            momenta, positions and Plumed.istep, where the positions and
            momenta corresponds to the last configuration in the previous
            simulation, while Plumed.istep is the number of timesteps
            performed previously. This can be done using
            ase.calculators.plumed.restart_from_trajectory.
        """

        # if atoms is None:
        #    raise TypeError('plumed calculator has to be defined with the \
        #                     object atoms inside.')

        self.istep = 0
        Calculator.__init__(self, atoms=atoms)

        self.input = input
        self.use_charge = use_charge
        self.update_charge = update_charge

        self.kT = kT
        self.restart = restart

        self._timestep = None
        self._stride = 1

        return

    @property
    def timestep(self):
        """"""

        return self._timestep

    @timestep.setter
    def timestep(self, timestep):
        """"""
        self._timestep = timestep

        return

    @property
    def stride(self):

        return self._stride

    @stride.setter
    def stride(self, stride):
        """"""
        self._stride = stride

        return self._stride

    def _prepare(
        self,
        natoms: int,
        input_lines: List[str],
        timestep: float,
        restart: bool,
        kT: float,
    ):
        """"""
        self.plumed = plumed.Plumed()

        # - basic configuration
        ps = 1000 * fs
        self.plumed.cmd("setMDEnergyUnits", mol / kJ)
        self.plumed.cmd("setMDLengthUnits", 1 / nm)
        self.plumed.cmd("setMDTimeUnits", 1 / ps)
        self.plumed.cmd("setMDChargeUnits", 1.0)
        self.plumed.cmd("setMDMassUnits", 1.0)
        self.plumed.cmd("setMDEngine", "ASE")

        self.plumed.cmd("setNatoms", natoms)
        self.plumed.cmd("setLogFile", os.path.join(self.directory, "plumed.out"))
        self.plumed.cmd("setTimestep", float(timestep))
        self.plumed.cmd("setRestart", restart)
        self.plumed.cmd("setKbT", float(kT))
        self.plumed.cmd("init")

        # - parse lines, update FILE and STRIDE
        input_lines, parsed_lines = copy.deepcopy(input_lines), []
        for line in input_lines:
            parsed_line = update_input_value(
                line, "FILE", self.directory, func=lambda x, y: os.path.join(y, x)
            )
            parsed_line = update_input_value(
                parsed_line, "STRIDE", self.stride, func=lambda x, y: str(y)
            )
            parsed_line = update_input_value(
                parsed_line, "PACE", self.stride, func=lambda x, y: str(y)
            )
            parsed_lines.append(parsed_line)

        for line in parsed_lines:
            self.plumed.cmd("readInputLine", line)

        # - save input lines
        inp_fpath = pathlib.Path(self.directory, "plumed.inp")
        with open(inp_fpath, "w") as fopen:
            fopen.write("\n".join(input_lines))

        return

    def _save_checkpoint(self, ckpt_wdir: pathlib.Path):
        """"""
        calc_wdir = pathlib.Path(self.directory)
        if (ckpt_wdir / calc_wdir.name).exists():
            shutil.rmtree(ckpt_wdir / calc_wdir.name)
        _ = shutil.copytree(calc_wdir, ckpt_wdir / calc_wdir.name)

        return

    def _load_checkpoint(
        self, ckpt_wdir: pathlib.Path, dst_wdir=None, start_step: int = 0
    ):
        """"""
        calc_wdir = list(ckpt_wdir.glob(f"*Plumed"))[0]
        if dst_wdir is None:
            dst_wdir = ckpt_wdir.parent
        else:
            dst_wdir = pathlib.Path(dst_wdir)
        _ = shutil.copytree(calc_wdir, dst_wdir / calc_wdir.name)

        if hasattr(self, "plumed"):
            self.plumed.finalize()
            delattr(self, "plumed")
        self.istep = int(start_step)
        self.input = ["RESTART"] + self.input
        self.restart = True

        return

    def calculate(
        self, atoms=None, properties=["energy", "forces"], system_changes=all_changes
    ):
        """"""
        Calculator.calculate(self, atoms, properties, system_changes)

        if self.timestep is None:
            raise RuntimeError(
                "Plumed needs a valid timestep set by an external class."
            )

        if not hasattr(self, "plumed"):
            self._prepare(len(atoms), self.input, self.timestep, self.restart, self.kT)

        energy_bias, forces_bias = self.compute_bias(
            self.atoms.get_positions(), self.istep
        )

        self.results["energy"], self.results["forces"] = energy_bias, forces_bias
        self.istep += 1

        return

    def compute_bias(self, pos, istep):
        """"""
        self.plumed.cmd("setStep", istep)

        # Box for functions with PBC in plumed
        if np.any(self.atoms.pbc):
            cell = self.atoms.get_cell(complete=True).array
            self.plumed.cmd("setBox", cell)

        self.plumed.cmd("setPositions", pos)
        self.plumed.cmd("setEnergy", 0.0)
        self.plumed.cmd("setMasses", self.atoms.get_masses())
        forces_bias = np.zeros((self.atoms.get_positions()).shape)
        self.plumed.cmd("setForces", forces_bias)
        virial = np.zeros((3, 3))
        self.plumed.cmd("setVirial", virial)
        self.plumed.cmd("prepareCalc")
        self.plumed.cmd("performCalc")
        energy_bias = np.zeros((1,))
        self.plumed.cmd("getBias", energy_bias)

        return [energy_bias, forces_bias]

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.plumed.finalize()


if __name__ == "__main__":
    ...
